/***********************************************************
*                                                          *
* A Bank Service Till (ATM) Simulation (Process-Oriented)  *
*                                                          *
* Source is:     16_01.CPP                                 *
* Created:       04 April, 1994                            *
* Last edited:   30 June, 1995                             *
*                                                          *
***********************************************************/

#include <iostream.h>

// extern "C" may be necessary on some compilers
// and won't do any harm on others
extern "C" 
{
#include <assert.h>
#include <stdlib.h>
#include <time.h>
}

// Maximum number of tries allowed to enter correct password
const int MAX_NUM_PASS_TRIES   = 3;

// Various "flags"
const int TRUE = 1;
const int FALSE = 0;

const int OUT_OF_PAPER                  = 10;
const int OUT_OF_CASH                   = 11;

const int CHECK_CARD                    = 12;
const int KEEP_CARD                     = 13;
const int EJECT_CARD                    = 14;
const int ORDER_CHEQUE_BOOK             = 15;
const int PRINT_BALANCE                 = 16;
const int PRINT_MINI_STATEMENT          = 17;
const int DISPENSE_CASH                 = 18;
const int QUIT                          = 19;

// Function prototypes 
void Check_cash_reserves(int&);
void Check_paper_reserves(int&);
void Check_supplies(int&, int&, int&, int&);
void Deliver_services(int&, int& ,int&, int&);
void Dispense_cash_to_customer(int&);
void Get_password(int&,int&,int&);
void Get_required_service(int&, int, int, int, int);
int  Hardware_fault_detected();
void Order_cheque_book_from_bank(int&);
void Print_customer_balance(int&);
void Print_customer_mini_statement(int&);
void Process_cash_card(int, int* = 0);  
void Process_required_service(int, int&, int&, int&, int&);
void Shutdown(void);


/***********************************************************
*                                                          *
*   CPPSPEC:  main()                                       *
*                                                          *
***********************************************************/
int main(void)
   {
   int cheque_OK,
       balance_OK,
       statement_OK,
       cash_OK;

   int valid_password_entered, valid_card_entered;
   int user_wants_to_quit, tries;

   // Simulate (up to) one customer in queue then
   // manual shut down 
   for (int customer = 1; customer <= 1; customer++)
      {
      // Welcome banner 
      cout << "----------------------------------\n";
      cout << "PROTOTYPE BANK AUTO TELLER MACHINE\n";
      cout << "----------------------------------\n\n";
      
      Check_supplies(cash_OK, balance_OK,
                     statement_OK, cheque_OK);
      
      Process_cash_card(CHECK_CARD, &valid_card_entered);
      
      if (valid_card_entered)
         {
         tries = 0; 
         user_wants_to_quit = FALSE;
         Get_password(valid_password_entered,
                      user_wants_to_quit, tries); 
         
         if ((valid_password_entered)
              && (!user_wants_to_quit))
            {
            // Keep delivering services till user says quit
            // (or hardware fault) 
            Deliver_services(cheque_OK, balance_OK,
                             statement_OK, cash_OK);
            Process_cash_card(EJECT_CARD);
            } 
         else
            {
            // No valid password 
            if (user_wants_to_quit)
               {
               // Polite ejection 
               Process_cash_card(EJECT_CARD);
               }  
            }   
         }
      } 
   
   // Simulate manual shutdown
   Shutdown();
   return 0;
   }


/***********************************************************
*                                                          *
*   CPPSPEC:  Shutdown()                                   *
*                                                          *
*   OVERVIEW: Exit message, and exit program.              *
*                                                          *
***********************************************************/
void Shutdown(void)
   {
   cout << "Please use another machine.\n";
   cout << "*** SIMULATING MACHINE SHUTDOWN ***\n";
   exit(1);   
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Check_supplies()                             *
*                                                          *
*   OVERVIEW: Check cash and paper reserves &c.            *
*                                                          *
***********************************************************/
void Check_supplies(int&  cash_OK_REF,
                    int&  balance_OK_REF,
                    int&  statement_OK_REF,
                    int&  cheque_OK_REF)
   {

   cout << "Next customer please wait...\n\n";

   Check_cash_reserves(cash_OK_REF);  

   int paper_level_OK; // Is paper level okay? (flag)
   Check_paper_reserves(paper_level_OK);

   // Assuming all supplies are OK
   cash_OK_REF = TRUE;
   balance_OK_REF = TRUE;
   statement_OK_REF = TRUE;
   cheque_OK_REF = TRUE;
   }


/***********************************************************
*                                                          *
*   CPPSPEC:  Hardware_fault_detected()                    *
*                                                          *
*   OVERVIEW: Make sure hardware is not damaged.           *
*                                                          *
***********************************************************/
int Hardware_fault_detected()
   {
   // Assume no hardware fault.
   return 0;
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Check_cash_reserves()                        *
*                                                          *
*   OVERVIEW: Check cash level.                            *
*                                                          *
***********************************************************/
void Check_cash_reserves(int& cash_OK_REF)
   {
   // Assume cash level OK
   cash_OK_REF = TRUE;
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Check_paper_reserves()                       *
*                                                          *
*   OVERVIEW: Check that there is paper to print on.       *
*                                                          *
***********************************************************/
void Check_paper_reserves(int& paper_level_OK_REF)
   {
   // Assume paper level OK
   paper_level_OK_REF = TRUE;
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Process_cash_card()                          *
*                                                          *
*   OVERVIEW: Validate, eject or retain cash card.         *
*                                                          *
***********************************************************/
void Process_cash_card(int required_op, int* card_OK_PTR)
   { 
   /* NOTE: Process_cash_card() is a function with a
      default argument.  Can call as:        
   
      Process_cash_card(CHECK_CARD, card_OK);
      or
      Process_cash_card(KEEP_CARD);
      Process_cash_card(EJECT_CARD);  */ 
  
   switch (required_op)
      {
      case CHECK_CARD:
         // Using "extra" argument 
         cout << "Please insert your card.\n";
         cout << "*** SIMULATING CARD INSERTION ***\n";
         *card_OK_PTR = TRUE; // Assume card OK
         cout << "Thank you.\n\n";
         // If NOT OK call Process_cash_card(EJECT_CARD);
         break;
                  
      case KEEP_CARD:
         cout << "Password incorrect. Card retained.\n";
         cout << "Please contact your branch.\n";
         cout << "*** SIMULATING SWALLOWING CARD ***\n\n";
         break;
         
      case EJECT_CARD:
         cout << "Thank you for using this machine.\n";
         cout << "*** SIMULATING EJECTING CARD ***\n\n";
         
         break;  
      }
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Get_password()                               *
*                                                          *
*   OVERVIEW: Get and validate user's password             *
*                                                          *
***********************************************************/
void Get_password(int& password_OK_REF,
                  int& want_to_quit_REF,
                  int& tries_REF)
   {
   int password;

   cout << "Please enter your password : ";
   cin >> password;
   
   cout << "*** SIMULATING CALL TO BANK COMPUTER ***\n";
   password_OK_REF = TRUE;
   cout << "*** SIMULATING CORRECT PASSWORD ***\n";
   tries_REF = 1;  // Assume password correct first time
   cout << "Thank you.\n\n";

   // Assume user does not want to quit
   want_to_quit_REF = FALSE; 
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Deliver_services()                           *
*                                                          *
*   OVERVIEW: Ask user for reqd. service and deliver it.   *
*                                                          *
***********************************************************/
void Deliver_services(int& cheque_OK_REF,
                      int& balance_OK_REF,
                      int& statement_OK_REF,
                      int& cash_OK_REF)
   {
   int user_choice;
   
   Get_required_service(user_choice, cheque_OK_REF,
       balance_OK_REF, statement_OK_REF, cash_OK_REF);

   while (user_choice != QUIT)
      {
      Process_required_service(user_choice, cheque_OK_REF,
         balance_OK_REF, statement_OK_REF, cash_OK_REF);
      Get_required_service(user_choice, cheque_OK_REF,
         balance_OK_REF, statement_OK_REF, cash_OK_REF);
      } 
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Get_required_service()                       *
*                                                          *
*   OVERVIEW: Find out what service user requires.         *
*                                                          *
***********************************************************/
void Get_required_service(int& service_REF,
                          int  cheque_OK,
                          int  balance_OK,
                          int  statement_OK,
                          int  cash_OK)
   {
   // Ignoring cheque_OK, etc: let user order
   // as many cheque books, etc, as she likes
   cout << "Which service do you require: \n";
   cout << "Quit              -  Type 1\n";
   cout << "Order cheque book -  Type 2\n";
   cout << "Print balance     -  Type 3\n";
   cout << "Print statement   -  Type 4\n";
   cout << "Cash              -  Type 5\n";

   int service;

   do {
      cout << "Please enter the appropriate number : "; 
      cin >> service;
      } while ((service > 5) || (service < 1));
   
   cout << "Thank you.\n\n";
   
   switch (service)
      {
      case 1: service_REF = QUIT; break;
      case 2: service_REF = ORDER_CHEQUE_BOOK; break; 
      case 3: service_REF = PRINT_BALANCE; break;
      case 4: service_REF = PRINT_MINI_STATEMENT; break;
      case 5: service_REF = DISPENSE_CASH; break;
      }
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Process_required_service()                   *
*                                                          *
*   OVERVIEW: Perform user's requested service.            *
*                                                          *
***********************************************************/
void Process_required_service(int  user_choice,
                              int& cheque_OK_REF, 
                              int& balance_OK_REF, 
                              int& statement_OK_REF, 
                              int& cash_OK_REF)
   {
   switch (user_choice)
      {
      case ORDER_CHEQUE_BOOK:
         Order_cheque_book_from_bank(cheque_OK_REF); 
      break;
      
      case PRINT_BALANCE:
         Print_customer_balance(balance_OK_REF);
      break;

      case PRINT_MINI_STATEMENT: 
         Print_customer_mini_statement(statement_OK_REF);
      break;

      case DISPENSE_CASH: 
         Dispense_cash_to_customer(cash_OK_REF);
      break;
      }
   }
   

/***********************************************************
*                                                          *
*   CPPSPEC:  Order_cheque_book_from_bank()                *
*                                                          *
*   OVERVIEW: Order cheque book.                           *
*                                                          *
***********************************************************/
void Order_cheque_book_from_bank(int& cheque_OK_REF)
   {
   cout << "*** SIMULATING ORDER FOR CHEQUE BOOK ***\n";
   cout << "Your cheque book will be posted to you.\n\n";
   }
   
/***********************************************************
*                                                          *
*   CPPSPEC:  Print_customer_balance()                     *
*                                                          *
*   OVERVIEW: Print customer's balance.                    *
*                                                          *
***********************************************************/
void Print_customer_balance(int& balance_OK_REF)
   {
   cout << "Your balance is: 99.99 (*** SIMULATED ***)\n\n";
   }
   
/***********************************************************
*                                                          *
*   CPPSPEC:  Print_customer_mini_statement()              *
*                                                          *
*   OVERVIEW: Print mini statement.                        *
*                                                          *
***********************************************************/
void Print_customer_mini_statement(int& statement_OK_REF)
   {
   cout << "*** SIMULATING PRINTING OF MINI STATEMENT ***";
   cout << "\nPlease take your mini statement.\n\n";
   }
   
/***********************************************************
*                                                          *
*   CPPSPEC:  Dispense_cash_to_customer()                  *
*                                                          *
*   OVERVIEW: Deliver cash.                                *
*                                                          *
***********************************************************/
void Dispense_cash_to_customer(int& cash_OK_REF)
   {
   cout << "*** SIMULATING CASH DELIVERY ***\n";
   cout << "Please take your cash.\n\n";
   }

/***********************************************************
*                 *** END OF PROGRAM ***                   *
***********************************************************/
