/***********************************************************
*                                                          *
* A Bank Service Till (ATM) Simulation (Process-Oriented)  *
*                                                          *
* Source is:     16_03s.CPP                                *
* Created:       04 Apr 1994                               *
* Last edited:   11 May 1997                               *
*                                                          *
***********************************************************/

//
// NOTE: This is a version of file 16_03a, in ISO C++
// See the main 'Read me' file on the CD for details
//

#include <cassert>
#include <cstdlib>
#include <ctime>
#include <iostream>

using namespace std;

// Maximum number of tries allowed to enter correct password
const int MAX_NUM_PASS_TRIES   = 3;

//
// Various "flags"
// Note: it would be better to use enumerated types here
// but we haven't yet covered this topic...
//
const int TRUE = 1;
const int FALSE = 0;

const int OUT_OF_PAPER                  = 10;
const int OUT_OF_CASH                   = 11;

const int CHECK_CARD                    = 12;
const int KEEP_CARD                     = 13;
const int EJECT_CARD                    = 14;
const int ORDER_CHEQUE_BOOK             = 15;
const int PRINT_BALANCE                 = 16;
const int PRINT_MINI_STATEMENT          = 17;
const int DISPENSE_CASH                 = 18;
const int QUIT                          = 19;

// Function prototypes 
void Check_Cash_Reserves(int&);
void Check_Paper_Reserves(int&);
void Deliver_Services(int&,int&,int&,int&);
void Dispense_Cash_To_Customer(int&);
void Display_Welcome_Screen(const int,const int);
void Get_Password(int&,int&,int&);
void Get_Required_Service(int&,const int,const int,const int,const int);
int  Hardware_Fault_Detected();
void Order_Cheque_Book_From_Bank(int&);
void Prepare_For_Next_Customer(int&,int&,int&,int&);
void Print_Customer_Balance(int&);
void Print_Customer_Mini_Statement(int&);
int  Process_Cash_Card(const int);  
void Process_Required_Service(const int,int&,int&,int&,int&);
void Shutdown(void);


/***********************************************************
*                                                          *
*   FUNCTION: main()                                       *
*                                                          *
***********************************************************/
int main()
   {
   int Cheque_available,
       Balance_available,
       Statement_available,
       Cash_available;

   int Valid_password_entered = FALSE;
   int Valid_card_entered = FALSE;
   int User_wants_to_quit = TRUE;
   int Tries;

   // Set up non-repeatable random numbers for simulation 
   srand((unsigned)time(NULL));
 
   // Simulate (up to) five customers in queue then
   // manual shut down 
   for (int Customer = 1; Customer <= 5; Customer++)
      {
      Prepare_For_Next_Customer(Cash_available, Balance_available,
                                Statement_available, Cheque_available);
      
      Valid_card_entered = Process_Cash_Card(CHECK_CARD);

      if (Valid_card_entered)
         {
         Tries = 0; 
         User_wants_to_quit = FALSE;
         Get_Password(Valid_password_entered,
                      User_wants_to_quit, Tries); 

         if ((Valid_password_entered)
              && (!User_wants_to_quit))
            {
            // Keep delivering services till user says quit
            // (or hardware fault) 
            Deliver_Services(Cheque_available, Balance_available,
                             Statement_available, Cash_available);

            Process_Cash_Card(EJECT_CARD);
            } 
         else
            {
            // No valid password 
            if (User_wants_to_quit)
               {
               // Polite ejection 
               Process_Cash_Card(EJECT_CARD);
               }  
            }   
         }
      } 

   // Simulate manual shutdown
   Shutdown();

   return EXIT_SUCCESS;
   }


/***********************************************************
*                                                          *
*   FUNCTION: Shutdown()                                   *
*                                                          *
*   OVERVIEW: Exit message, and exit program.              *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Shutdown(void)
   {
   // Pre - none

   cout << "Please use another machine.\n";
   cerr << "*** SIMULATING MACHINE SHUTDOWN ***\n";
   exit(1);   

   // Post - none(!)
   } 


/***********************************************************
*                                                          *
*   FUNCTION: Hardware_Fault_Detected()                    *
*                                                          *
*   OVERVIEW: Make sure hardware is not damaged.           *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  TRUE (1) if fault, FALSE (0) otherwise.      *
*                                                          *
***********************************************************/
int Hardware_Fault_Detected()
   {
   // Pre - none

   // Simulating a hardware fault with a prob. of 1/100  
   if ((rand() % 100) == 0)
      {
      cerr << "*** SIMULATING HARDWARE FAULT ***\n";
      return 1;
      }
   else
      {
      return 0;
      }

   // Post - none
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Check_Cash_Reserves()                        *
*                                                          *
*   OVERVIEW: Check cash level.                            *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     Got valid data.                              *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Check_Cash_Reserves(int& rCash_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   // Simulating a 1/6 chance of a cash shortage
   if ((rand() % 6)==0)
      {
      rCash_available = FALSE;
      cerr << "*** SIMULATING LOW CASH LEVEL ***\n";
      }      
   else
      {
      rCash_available = TRUE;
      }

   // Post - got valid data
   assert((rCash_available == FALSE)
         || (rCash_available == TRUE));
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Check_Paper_Reserves()                       *
*                                                          *
*   OVERVIEW: Check that there is paper to print on.       *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     Got valid data.                              *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Check_Paper_Reserves(int& rPaper_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   // Simulating a 1/6 chance of a paper shortage
   if ((rand() % 6)==0)
      {
      rPaper_available = FALSE;
      cerr << "*** SIMULATING LOW PAPER LEVEL ***\n";
      }      
   else
      {
      rPaper_available = TRUE;
      }
 
   // Post - got valid data
   assert((rPaper_available == FALSE)
         || (rPaper_available == TRUE));
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Process_Cash_Card()                          *
*                                                          *
*   OVERVIEW: Validate, eject or retain cash card.         *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Valid operation requested.                   * 
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  When validating, returns TRUE or FALSE       *
*                                                          *
***********************************************************/
int Process_Cash_Card(const int Required_op)
   { 
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }
   // Pre - valid operation requested
   assert((Required_op == CHECK_CARD) ||
          (Required_op == KEEP_CARD) ||
          (Required_op == EJECT_CARD));

   int Card_ok = FALSE;

   switch (Required_op)
      {
      case CHECK_CARD:
         // Validate user's card 
         Card_ok = (((rand() % 6) != 0) ? (TRUE) : (FALSE));
         
         if (Card_ok)
            {
            cerr <<
            "*** SIMULATING CORRECT CARD INSERTED ***\n";
            cout << "Thank you.\n\n";
            }
         else
            {
            cerr <<
            "*** SIMULATING *INCORRECT* CARD INSERTED ***\n"; 
            cout << "Incorrect card.\n";
            Process_Cash_Card(EJECT_CARD);
            }
         
         return Card_ok;
                  
      case KEEP_CARD:
         cout << "Password incorrect. Card retained.\n";
         cout << "Please contact your branch.\n";
         cerr << "*** SIMULATING SWALLOWING CARD ***\n\n";
         
         return 0;
         
      case EJECT_CARD:
         cout << "Thank you for using this machine.\n";
         cout << "Please take your card.\n";
         cerr << "*** SIMULATING EJECTING CARD ***\n\n";
         
         return 0;
         
      default:
         cerr << "*** INVALID OPERATION ***\n\n";
         return 0;
      }

   // Post - none
   }

 
/***********************************************************
*                                                          *
*   FUNCTION: Display_Welcome_Screen()                     *
*                                                          *
*   OVERVIEW: Display appropriate welcome banner.          *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     None                                         *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_Welcome_Screen(const int CASH_AVAILABLE, 
                            const int PAPER_AVAILABLE)
   {
   // Welcome banner 
   cout << "----------------------------------\n";
   cout << "PROTOTYPE BANK AUTO TELLER MACHINE\n";
   cout << "----------------------------------\n\n";

   if (!CASH_AVAILABLE)
      {
      cout << "Sorry - this ATM has no cash available\n\n";
      }

   if (!PAPER_AVAILABLE)
      {
      cout << "Sorry - this ATM cannot print receipts,\n"
           << "        balances or mini-statements.\n\n";
      }

   cout << "Please insert your card...\n";

   cout << "*** PLEASE HIT RETURN TO SIMULATE CARD INSERTION ***\n";

   char Tmp[80];
   cin.getline(Tmp, 80);
   }

/***********************************************************
*                                                          *
*   FUNCTION: Get_Password()                               *
*                                                          *
*   OVERVIEW: Get and validate user's password             *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     Got valid data.                              *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Get_Password(int& rPassword_correct,
                  int& rWant_to_quit,
                  int& rTries)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   int  Password;
   char Password_string[81];
   
   rTries++;                                             
   cout << "Please enter your password"
        << "\n(or 9999 to simulate QUIT button): ";
   cin.getline(Password_string, sizeof(Password_string));
   
   Password = atoi(Password_string);

   if (!(Password == 9999))
      {
      // User did not ask to quit

      // Simulate call to bank computer 
      cerr << "*** SIMULATING CALL TO BANK COMPUTER ***\n";

      // Simulating one single valid password... 
      // In reality, send account number and password
      // to bank computer 
      // Computer returns 'YES' or 'NO'.

      if (Password == 1234)
         {
         rPassword_correct = TRUE;
         }
      else
         {
         rPassword_correct = FALSE;
         }
   
      if (!(rPassword_correct)) 
         {
         if (rTries < MAX_NUM_PASS_TRIES)
            {
            cout << "Password incorrect.\n";
            // Recursive call to Get_Password()          
            Get_Password(rPassword_correct, rWant_to_quit, rTries);
            }
         else
            {
            // Too many tries - could be a stolen card?  
            Process_Cash_Card(KEEP_CARD);
            }
         }
      else
         {      
         cout << "Thank you.\n\n";
         }
      }
    else
      {
      // User wants to quit 
      rWant_to_quit = TRUE; 
      }     

   // Post - got valid data
   assert((rPassword_correct == FALSE)
         || (rPassword_correct == TRUE));
   assert((rWant_to_quit == FALSE)
         || (rWant_to_quit == TRUE));
   assert(rTries <= MAX_NUM_PASS_TRIES);
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Deliver_Services()                           *
*                                                          *
*   OVERVIEW: Ask user for reqd. service and deliver it.   *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid data.                              *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Deliver_Services(int& rCheque_available,
                      int& rBalance_available,
                      int& rStatement_available,
                      int& rCash_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   // Pre - got valid data
   assert((rCash_available == FALSE)
         || (rCash_available == TRUE));
   assert((rCheque_available == FALSE)
         || (rCheque_available == TRUE));
   assert((rBalance_available == FALSE)
         || (rBalance_available == TRUE));
   assert((rStatement_available == FALSE)
         || (rStatement_available == TRUE));

   int User_choice;
   
   Get_Required_Service(User_choice, rCheque_available,
       rBalance_available, rStatement_available, rCash_available);

   while (User_choice != QUIT)
      {
      Process_Required_Service(User_choice, rCheque_available,
         rBalance_available, rStatement_available, rCash_available);
      Get_Required_Service(User_choice, rCheque_available,
         rBalance_available, rStatement_available, rCash_available);
      } 

   // Post - none
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Get_Required_Service()                       *
*                                                          *
*   OVERVIEW: Find out what service user requires.         *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     Got valid choice.                            *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Get_Required_Service(int&      rService,
                          const int CHEQUE_AVAILABLE,
                          const int BALANCE_AVAILABLE,
                          const int STATEMENT_AVAILABLE,
                          const int CASH_AVAILABLE)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   int Service;

   cout << "Which service do you require: \n";
                     cout << "Quit              -  Type 1\n";
   if (CHEQUE_AVAILABLE)    cout << "Order cheque book -  Type 2\n";
   if (BALANCE_AVAILABLE)   cout << "Print balance     -  Type 3\n";
   if (STATEMENT_AVAILABLE) cout << "Print statement   -  Type 4\n";
   if (CASH_AVAILABLE)      cout << "Cash              -  Type 5\n";

   char Service_string[81];

   do {
      cout << "Please enter the appropriate number : "; 
      cin.getline(Service_string, sizeof(Service_string));
      Service = atoi(Service_string);
      } while ((Service > 5) 
               || (Service < 1)
               || ((Service == 2) && (!CHEQUE_AVAILABLE))
               || ((Service == 3) && (!BALANCE_AVAILABLE))
               || ((Service == 4) && (!STATEMENT_AVAILABLE))
               || ((Service == 5) && (!CASH_AVAILABLE)));
   
   cout << "Thank you.\n\n";
   
   switch (Service)
      {
      case 1: rService = QUIT; break;
      case 2: rService = ORDER_CHEQUE_BOOK; break; 
      case 3: rService = PRINT_BALANCE; break;
      case 4: rService = PRINT_MINI_STATEMENT; break;
      case 5: rService = DISPENSE_CASH; break;
      }

   // Post - got valid choice
   assert((rService == QUIT) ||
          (rService == ORDER_CHEQUE_BOOK) ||
          (rService == PRINT_BALANCE) ||
          (rService == PRINT_MINI_STATEMENT) ||
          (rService == DISPENSE_CASH));
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Process_Required_Service()                   *
*                                                          *
*   OVERVIEW: Perform user's requested service.            *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid choice of service.                 *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Process_Required_Service(const int  User_choice,
                                    int& rCheque_available, 
                                    int& rBalance_available, 
                                    int& rStatement_available, 
                                    int& rCash_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   // Pre - got valid choice
   assert((User_choice == QUIT) ||
          (User_choice == ORDER_CHEQUE_BOOK) ||
          (User_choice == PRINT_BALANCE) ||
          (User_choice == PRINT_MINI_STATEMENT) ||
          (User_choice == DISPENSE_CASH));

   switch (User_choice)
      {
      case ORDER_CHEQUE_BOOK:
         Order_Cheque_Book_From_Bank(rCheque_available); 
      break;
      
      case PRINT_BALANCE:
         Print_Customer_Balance(rBalance_available);
      break;

      case PRINT_MINI_STATEMENT: 
         Print_Customer_Mini_Statement(rStatement_available);
      break;

      case DISPENSE_CASH: 
         Dispense_Cash_To_Customer(rCash_available);
      break;
      }

   // Post - none
   }
   

/***********************************************************
*                                                          *
*   FUNCTION: Order_Cheque_Book_From_Bank()                *
*                                                          *
*   OVERVIEW: Order cheque book.                           *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid data.                              *
*   POST:     Returning valid data.                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Order_Cheque_Book_From_Bank(int& rCheque_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }
   // Pre - got valid data
   assert((rCheque_available == TRUE));

   cerr << "*** SIMULATING ORDER FOR CHEQUE BOOK ***\n";
   cout << "Your cheque book will be posted to you.\n\n";

   rCheque_available = FALSE;
 
   // Post - returning valid data
   assert((rCheque_available == FALSE));
  }

 
/***********************************************************
*                                                          *
*   FUNCTION: Prepare_For_Next_Customer()                  *
*                                                          *
*   OVERVIEW: Check cash and paper reserves &c.            *
*                                                          *
*   PRE:      No hardware faults.                          *
*   POST:     Got valid data.                              *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Prepare_For_Next_Customer(int&  rCash_available,
                    int&  rBalance_available,
                    int&  rStatement_available,
                    int&  rCheque_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }

   int Paper_level_OK; // Is paper level okay? (flag)

   cout << "Next customer please wait...\n\n";

   Check_Cash_Reserves(rCash_available);  
   Check_Paper_Reserves(Paper_level_OK);  

   if (!Paper_level_OK)
      {
      rBalance_available = FALSE;
      rStatement_available = FALSE;
      }
   else   
      {
      // Reset these for new customer 
      rBalance_available = TRUE;
      rStatement_available = TRUE;
      }

   // Reset flag for new customer (always available)
   rCheque_available = TRUE;

   // Now display welcome banner
   Display_Welcome_Screen(rCash_available, Paper_level_OK);

   // Post - got valid data
   assert((rCash_available == FALSE)
         || (rCash_available == TRUE));
   assert((rBalance_available == FALSE)
         || (rBalance_available == TRUE));
   assert((rStatement_available == FALSE)
         || (rStatement_available == TRUE));
   assert((rCheque_available == FALSE)
         || (rCheque_available == TRUE));
   }

  
/***********************************************************
*                                                          *
*   FUNCTION: Print_Customer_Balance()                     *
*                                                          *
*   OVERVIEW: Print customer's balance.                    *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid data.                              *
*   POST:     Returning valid data.                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Print_Customer_Balance(int& rBalance_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }
   // Pre - got valid data
   assert((rBalance_available == TRUE));

   cout << "Your balance is: 99.99 (*** SIMULATED ***)\n\n";

   rBalance_available = FALSE;

   // Post - returning valid data
   assert((rBalance_available == FALSE));
   }
   
/***********************************************************
*                                                          *
*   FUNCTION: Print_Customer_Mini_Statement()              *
*                                                          *
*   OVERVIEW: Print mini statement.                        *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid data.                              *
*   POST:     Returning valid data.                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Print_Customer_Mini_Statement(int& rStatement_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }
   // Pre - got valid data
   assert((rStatement_available == TRUE));

   cerr << "*** SIMULATING PRINTING OF MINI STATEMENT ***";
   cout << "\nPlease take your mini statement.\n\n";

   rStatement_available = FALSE;

   // Post - returning valid data
   assert((rStatement_available == FALSE));
   }
   
/***********************************************************
*                                                          *
*   FUNCTION: Dispense_Cash_To_Customer()                  *
*                                                          *
*   OVERVIEW: Deliver cash.                                *
*                                                          *
*   PRE:      No hardware faults.                          *
*             Got valid data.                              *
*   POST:     Returning valid data.                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Dispense_Cash_To_Customer(int& rCash_available)
   {
   // Pre - no hardware faults
   if (Hardware_Fault_Detected()) { Shutdown(); }
   // Pre - got valid data
   assert((rCash_available == TRUE));

   cerr << "*** SIMULATING CASH DELIVERY ***\n";
   cout << "Please take your cash.\n\n";

   rCash_available = FALSE;

   // Post - returning valid data
   assert((rCash_available == FALSE));
   }

/***********************************************************
*                 *** END OF PROGRAM ***                   *
***********************************************************/




