// Playing with structs, unions and enums
// Source is: 17_04.CPP

#include <iostream.h>
#include <stdlib.h>
#include <iomanip.h>

// "Data prototypes"
enum eBoolean{FALSE, TRUE};

struct sEmployee
   {
   char     surname[20];
   eBoolean hourly_paid;
   union // Anonymous
      {
      // Hourly-paid workers
      // Need hours worked, and rate/hour
      struct // Anonymous
         {
         int   hours;
         float hourly_rate;
         } hourly;
      // Salaried workers
      // Just need monthly salary   
      float monthly_salary;
      } pay;
   };

// Function prototypes
int Get_Pay_Data(sEmployee* const);
int Print_Gross_Pay(const sEmployee);

int main ()
   {
   // Local variables
   int       emp_num;
   sEmployee payroll[2]; // The payroll "database"
 
   // Print welcome message
   cout << "Basic Payroll Program\n";
   cout << "---------------------\n\n";
 
   // Add some payroll data 
   for (emp_num = 0; emp_num <= 1; emp_num++)
      {
      if (Get_Pay_Data(&payroll[emp_num]))
         {
         cout << "FATAL ERROR: Function Get_Pay_Data()\n";
         exit (1);
         }
      }

   // Print table heading 
   cout << "GROSS PAY\n";
   cout << "---------\n";
   // Display the payroll data
   for (emp_num = 0; emp_num <= 1; emp_num++)
      {
      if (Print_Gross_Pay(payroll[emp_num]))
         {
         cout << "FATAL ERROR: Function Print_Gross_Pay()\n";
         exit (1);
         }
      }

   cout << "\nPay calculations complete.";

   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION:  Get_Pay_Data                                *
*                                                          *
*   OVERVIEW:  Get pay data for specified employee.        *
*                                                          *
***********************************************************/ 
int Get_Pay_Data(sEmployee* const employee_PTR)
   {
   char text_str[10];

   cout << "EMPLOYEE SURNAME : ";
   cin.getline(employee_PTR->surname, 
               sizeof(employee_PTR->surname));

   cout << "Is employee " << employee_PTR->surname << "\n";
   cout << "* Paid by the HOUR (Type 1 <RETURN>)\n";
   cout << "* Paid by SALARY   (Type 0 <RETURN>)\n";
   cout << " ? : ";

   // NO error checking...
   cin.getline(text_str, sizeof(text_str));  

   if (atoi(text_str))
      {
      employee_PTR->hourly_paid = TRUE;
      }
   else
      {
      employee_PTR->hourly_paid = FALSE;
      }

   if(employee_PTR->hourly_paid == TRUE)
      {
      // Hourly-paid employee
      cout << "How many hours were worked (whole number) : ";
      cin.getline(text_str, sizeof(text_str));
      employee_PTR->pay.hourly.hours = atoi(text_str);
   
      cout << "What is the hourly rate (.pp) : ";
      cin.getline(text_str, sizeof(text_str));
      employee_PTR->pay.hourly.hourly_rate = (float)atof(text_str);
      }
    else
      {
      // Salaried employee
      cout << "What is the monthly salary (.pp) : ";
      cin.getline(text_str, sizeof(text_str));
      employee_PTR->pay.monthly_salary = (float)atof(text_str);
      }
   cout << "\n";  

   return 0;
   }
                     
/***********************************************************
*                                                          *
*   FUNCTION:  Print_Gross_Pay                             *
*                                                          *
*   OVERVIEW:  Display gross pay for specified employee.   *
*                                                          *
***********************************************************/                       
int Print_Gross_Pay(const sEmployee EMPLOYEE)
   {
   float gross_pay;

   if (EMPLOYEE.hourly_paid == TRUE)
      {
      gross_pay = EMPLOYEE.pay.hourly.hours
                  * EMPLOYEE.pay.hourly.hourly_rate;
      }
   else
      {
      gross_pay = EMPLOYEE.pay.monthly_salary;
      }
   
   cout << setiosflags(ios::fixed)
        << setprecision(2);
   cout << EMPLOYEE.surname << " " << gross_pay << "\n";

   return 0;
   }
