// Playing with a Linked List
// Source is: 18_03.CPP

#include <iostream.h>

// sElement is the SELF REFERENTIAL structure at
// the heart of the list.
struct sElement
   {
   // The actual data (just one char in this example)
   char data;
   // The ptr (link) to the next list entry     
   sElement* next_PTR;
   } ;

void  Insert_element(sElement**, char);
char  Delete_element(sElement**, char);
int   Is_empty(sElement*);
void  Display_list(sElement*);
void  Help();

int main()
   {
   sElement* start_PTR = NULL;

   int choice = 0;
   char item;

   cout << "LINKED LIST DEMONSTRATION\n";
   cout << "-------------------------";

   Display_list(start_PTR);
   Help();

   while (choice != 3)
      {
      cout << "\nInsert (1), Delete (2) or Quit (3) ? ";
      cin >> choice;
   
      switch (choice)
         {
         case 1:
            cout << "\nEnter a character for insertion: ";
            cin >> item;
            Insert_element(&start_PTR, item);
            Display_list(start_PTR);
            break;
         
         case 2:
            if (!Is_empty(start_PTR))
               {
               cout << "\nEnter character to be deleted: ";
               cin >> item;
         
               if (Delete_element(&start_PTR, item))
                  {
                  cout << item << " deleted.\n";
                  Display_list(start_PTR);
                  }
               else
                  {
                  cout << item << " not found.\n\n";
                  }
               }     
            else
               {
               cout << "List is empty.\n\n";
               }
            break;
         case 3:
            cout << "\nThank you for using this program.\n";
            break;
         default:
            cout << "\n*** INVALID OPTION ***\n";
            Help();
         }
      }

   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Help()                                       *
*                                                          *
*   OVERVIEW: Tell the user how to use the program.        *
*                                                          *
***********************************************************/
void Help()
   {
   cout << "\n"
        << "Type 1 to insert an element into the list.\n"
        << "     2 to delete an element from the list.\n"
        << "     3 to QUIT.\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Insert_element()                             *
*                                                          *
*   OVERVIEW: Insert a new element into the list           *
*.            (in alphabetical order).                     *
*                                                          *
***********************************************************/
// NOTE: E_PTR_PTR is a pointer to a pointer
void Insert_element(sElement** E_PTR_PTR, char value)
   {
   sElement* new_PTR;
   sElement* last_PTR;
   sElement* this_PTR;

   if (new_PTR = new sElement)
      {
      new_PTR->data = value;
      new_PTR->next_PTR = NULL;

      last_PTR = NULL;
      this_PTR = *E_PTR_PTR;

      while ((this_PTR != NULL) && (value > this_PTR->data))
         {
         last_PTR = this_PTR;
         this_PTR = (this_PTR)->next_PTR;
         }

      if (last_PTR == NULL)
         {
         new_PTR->next_PTR = *E_PTR_PTR;
         *E_PTR_PTR = new_PTR;
         }
      else
         {
         last_PTR->next_PTR = new_PTR;
         new_PTR->next_PTR = this_PTR;
         }
      }
   else
      {
      cout << value << " not inserted. No memory available.\n";
      }
   }

/***********************************************************
*                                                          *
*   FUNCTION: Delete_element()                             *
*                                                          *
*   OVERVIEW: Deletes an element from the list.            *
*                                                          *
***********************************************************/
// NOTE: E_PTR_PTR is a pointer to a pointer
char Delete_element(sElement** E_PTR_PTR, char value)
   {
   // Compare this with the definition in Insert_element...
   sElement *last_PTR, *this_PTR, *temp_PTR; 

   if (value == (*E_PTR_PTR)->data)
      {
      temp_PTR = *E_PTR_PTR;
      *E_PTR_PTR = (*E_PTR_PTR)->next_PTR;
      delete temp_PTR;
      temp_PTR = 0;
      return value;
      }
   else
      {
      last_PTR = *E_PTR_PTR;
      this_PTR = (*E_PTR_PTR)->next_PTR;
   
      while ((this_PTR != NULL) && (this_PTR->data != value))
         {
         last_PTR = this_PTR;
         this_PTR = this_PTR->next_PTR;
         }
   
      if (this_PTR)
         {
         temp_PTR = this_PTR;
         last_PTR->next_PTR = this_PTR->next_PTR;
         delete temp_PTR;
         temp_PTR = 0;
         return value;
         }
      }

   return '\0';
   }

/***********************************************************
*                                                          *
*   FUNCTION: Is_empty()                                   *
*                                                          *
*   OVERVIEW: Returns 1 if the list is empty, 0 otherwise. *
*                                                          *
***********************************************************/
int Is_empty(sElement* E_PTR)
   {
   if (E_PTR == NULL)
      {
      return 1;
      }
   else
      {
      return 0;
      }
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_list()                               *
*                                                          *
*   OVERVIEW: Display all the list elements on the screen. *
*                                                          *
***********************************************************/
void Display_list(sElement* this_PTR)
   {
   if (this_PTR == NULL)
      {
      cout << "\nList starts at address " 
           << this_PTR << endl;
      cout << "(List is empty)\n";
      cout << "List ends at address " 
           << this_PTR << endl;
      }
   else
      {
      cout << "\nList starts at address " 
           << this_PTR << endl;
   
      while (this_PTR != NULL)
         {
         cout << "Data " << (char) this_PTR->data << " -> ";
         cout << this_PTR->next_PTR << endl;
         this_PTR = this_PTR->next_PTR;
         }                                   
      
      cout << "List ends at address " 
           << this_PTR << endl;
      }
   }
/***********************************************************
*                 *** END OF PROGRAM ***                   *
***********************************************************/
