/***********************************************************
*                                                          *
*  A Simple (D-O) Stock Control System                     *
*                                                          *
*  Source is:     23_01.CPP                                *
*  Created:       11 Dec, 1994                             *
*  Last edited:   28 Jul, 1995                             *
*                                                          *
***********************************************************/

#include <fstream.h>
#include <iomanip.h>
#include <iostream.h>

// extern "C" may be necessary on some compilers
// and won't do any harm on others
extern "C" 
{
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
}

// Maximum number of different products in stock
const int MAX_STOCK = 1000;

// Maximum number of different suppliers 
const int MAX_SUPPLIER = 3;     

// If the stock level of a particular product falls 
// below LOW_STOCK_LEVEL, an order is placed for more 
const int LOW_STOCK_LEVEL = 50;   

struct sSupp
   {
   int   SuppNum;  
   char  SuppName[30];
   char  SuppAddr[30]; 
   };

struct sProd
   {
   int   ProdNum;  
   char  ProdName[30]; 
   int   ProdQuan;        //Number of items in stock
   int   ProdCost;        //Price of items - in PENCE.
   int   SuppNum;
   };

// Function declarations
void Browse_suppliers();
void Delete_record();
void Display_all_records();
void Display_low_stock_records();
void Insert_record();
int  Menu();
void Open_files();
void Update_record();

/***********************************************************
*                                                          *
*   FUNCTION: main()                                       *
*                                                          *
***********************************************************/
int main()
   {
   cout << "Prototype Stock Control System\n";
   cout << "------------------------------\n";

   Open_files();
      
   int process; 
   while ((process = Menu()) != 7) 
      {
      switch (process) 
         {
         case 1:
            Display_all_records();
            break;
         case 2:
            Display_low_stock_records();
            break;         
         case 3:
            Update_record();
            break;
         case 4:
            Insert_record();
            break;            
         case 5:
            Delete_record();
            break;
         case 6:
            Browse_suppliers();
          }
      }

   cout << "Finished.";
   
   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Menu()                                       *
*                                                          *
*   PURPOSE:  Displays user instructions                   *
*                                                          *
***********************************************************/
int Menu() 
   {
   // Local variables 
   int choice;
   char ip_STR[80];

   cout  << "\nMENU:\n"
         << "1  List all stock items.\n"
         << "2  List items with low stock level.\n"
         << "3  Update record for a particular item.\n"
         << "4  Insert record for a new item.\n"
         << "5  Delete record for an item.\n"
         << "6  Browse supplier details.\n"
         << "7  QUIT program.\n\n"
         << "? :";
         
   cin.getline(ip_STR, sizeof(ip_STR));    
   
   while ((!(sscanf(ip_STR, "%d", &choice) == 1)) 
         || (choice < 1) || (choice > 7))
      {
      cout << "\nOption must be in the range 1 - 7. "
           << "Please re-enter: ";
      cin.getline(ip_STR, sizeof(ip_STR));    
      }
   
   cout << endl;
   
   return choice;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_all_records()                        *
*                                                          *
*   PURPOSE:  Lists all non-empty records                  *
*                                                          *
***********************************************************/
void Display_all_records()
   {
   cout << "*** WILL DISPLAY ALL RECORDS ***\n";
   } 

/***********************************************************
*                                                          *
*   FUNCTION: Display_low_stock_records()                  *
*                                                          *
*   PURPOSE:  Lists all products with low stock levels.    *
*                                                          *
***********************************************************/
void Display_low_stock_records()
   {
   cout << "*** WILL DISPLAY LOW STOCK RECORDS ***\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Update_record()                              *
*                                                          *
*   PURPOSE:  Updates contents of a specified record.      *
*                                                          *
***********************************************************/
void Update_record()
   {
   cout << "*** WILL UPDATE RECORD ***\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Browse_suppliers()                           *
*                                                          *
*   PURPOSE:  List and edit available suppliers.           *
*                                                          *
***********************************************************/
void Browse_suppliers()
   {
   cout << "*** WILL ALLOW EDITING OF SUPPLIERS ***\n";
   }


/***********************************************************
*                                                          *
*   FUNCTION: Insert_record()                              *
*                                                          *
*   PURPOSE:  Allows a new record to be inserted in file.  *
*                                                          *
***********************************************************/
void Insert_record()
   {
   cout << "*** WILL INSERT RECORD ***\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Delete_record()                              *
*                                                          *
*   PURPOSE:  Allows deletion of a specified record.       *
*                                                          *
***********************************************************/
void Delete_record()
   {
   cout << "*** WILL DELETE RECORD ***\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Open_files()                                 *
*                                                          *
*   PURPOSE:  Open database.                               *
*                                                          *
***********************************************************/
void Open_files()
   {
   cout << "*** WILL OPEN FILES ***\n";
   }

/***********************************************************
*                  *** END OF PROGRAM ***                  *
***********************************************************/
