/***********************************************************
*                                                          *
*  A Simple (D-O) Stock Control System                     *
*                                                          *
*  Source is:     23_02a.CPP                               *
*  Created:       11 Dec, 1994                             *
*  Last edited:   12 May, 1997                             *
*                                                          *
***********************************************************/

//
// This version (23_02a.cpp) reflects recent changes to the 
// iostream library during the moves towards 'Standard C++'.  
// The program is compatible with most modern C++ 
// compilers.
//
// See also file 23_02s.cpp which is in 'Standard C++'.
//
// MJP, 12 May, 1997
//

#include <fstream.h>
#include <iomanip.h>
#include <iostream.h>

// extern "C" may be necessary on some compilers
// and won't do any harm on others
extern "C" 
{
#include <assert.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
}

// Maximum number of different products in stock
const int MAX_STOCK = 1000;

// Maximum number of different suppliers
const int MAX_SUPPLIER = 3;     

// If the stock level of a particular product falls 
// below LOW_STOCK_LEVEL, an order is placed for more 
const int LOW_STOCK_LEVEL = 50;   

struct sSupp
   {
   int   SuppNum;  
   char  SuppName[30];
   char  SuppAddr[30]; 
   };

struct sProd
   {
   int   ProdNum;  
   char  ProdName[30]; 
   int   ProdQuan;        //Number of items in stock
   int   ProdCost;        //Price of items - in PENCE.
   int   SuppNum;
   };

// Function declarations
int  Browse_suppliers(fstream&);
int  Choose_supplier(int&, fstream&);
void Delete_record(fstream&);
void Display_all_records(fstream&, fstream&);
void Display_low_stock_records(fstream&, fstream&);
void Display_money(const long);
void Display_record(const sProd, fstream&);
void Display_record_footer();
void Display_record_header();
void Fill_record(fstream&, fstream&, const int);
void Format_new_stock_file(fstream&);
void Format_new_supplier_file(fstream&);
void Get_all_data(fstream&,fstream&);
void Get_supplier_data(fstream&);
void Get_valid_ProdNum(int&);
void Insert_record(fstream&, fstream&);
int  Menu();
void Open_files(fstream&, fstream&);
void Type_enter_to_continue();
void Update_record(fstream&, fstream&);

/***********************************************************
*                                                          *
*   FUNCTION: main()                                       *
*                                                          *
***********************************************************/
int main()
   {
   // Local variables 
   fstream stock_FPTR; // File pointer: stock data file
   fstream supp_FPTR;  // File pointer: supplier data file

   cout << "Prototype Stock Control System\n";
   cout << "------------------------------\n";

   Open_files(stock_FPTR, supp_FPTR);
      
   int process; 
   while ((process = Menu()) != 7) 
      {
      switch (process) 
         {
         case 1:
            Display_all_records(stock_FPTR, supp_FPTR);
            break;
         case 2:
            Display_low_stock_records(stock_FPTR, supp_FPTR);
            break;         
         case 3:
            Update_record(stock_FPTR, supp_FPTR);
            break;
         case 4:
            Insert_record(stock_FPTR, supp_FPTR);
            break;            
         case 5:
            Delete_record(stock_FPTR);
            break;
         case 6:
            Browse_suppliers(supp_FPTR);
            break;
         }
      }

   stock_FPTR.close();

   cout << "Finished.";
   
   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Type_enter_to_continue()                     *
*                                                          *
*   OVERVIEW: Waits for ENTER, then returns.               *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Type_enter_to_continue()
   {
   // Local variable 
   char ip_STR[81];

   cout << "\nType <ENTER> to continue.";
   cin.getline(ip_STR, sizeof(ip_STR));    
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_record()                             *
*                                                          *
*   PURPOSE:  Display (on screen) a single record.         *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_record(const sProd REC, fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointer
   assert(supp_FPTR_REF.is_open());

   // Local variable
   sSupp temp;
   
   cout.setf(ios::left);
   cout  << " " << setw(10) << REC.ProdNum 
         << setw(25) << REC.ProdName; 

   // Get supplier's name
   supp_FPTR_REF.seekg(REC.SuppNum*sizeof(sSupp), ios::beg);
   supp_FPTR_REF.read((char*) &temp, sizeof(sSupp));
   cout  << setw(15) << temp.SuppName;

   cout  << setw(10) << REC.ProdQuan; 
   Display_money(REC.ProdCost);  
   cout << "\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_record_header                        *
*                                                          *
*   PURPOSE:  Display header for subsequent record(s).     *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_record_header()
   {
   cout.setf(ios::left);
   cout << "+---------+------------------------"
        << "+--------------+---------+-------+\n";
   cout << setw(10) << " Record #"
        << setw(25) << " Product Name"
        << setw(15) << " Supplier" 
        << setw(10) << " Quantity"
        << setw(8)  << " Cost" << endl;
   cout << "+---------+------------------------"
        << "+--------------+---------+-------+\n";
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_record_footer()                      *
*                                                          *
*   PURPOSE:  Display footer for previous record(s).       *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_record_footer()
   {
   cout.setf(ios::left);
   cout << "+---------+------------------------"
        << "+--------------+---------+-------+\n";
   }
  
/***********************************************************
*                                                          *
*   FUNCTION: Fill_record()                                *
*                                                          *
*   PURPOSE:  Prompt user for data (single record).        *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Fill_record(fstream& stock_FPTR_REF,
                 fstream& supp_FPTR_REF,
                 const int PROD_NUM)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   char ip_STR[80]; 
   sProd temp;
   int SuppNumber;

   temp.ProdNum = PROD_NUM;

   cout << "PRODUCT NAME : ";
   // Need to use getline because of possible white space..
   cin.getline(temp.ProdName, sizeof(temp.ProdName));
   while (!temp.ProdName[0])
       {
       // User isn't able to enter blank field
       cout << "\rPRODUCT NAME : ";
       cin.getline(temp.ProdName, sizeof(temp.ProdName));
       } 
      
   cout << "QUANTITY     : ";
   cin.getline(ip_STR, sizeof(ip_STR));
   while (!ip_STR[0])
      {
      // User isn't able to enter blank field
      cout << "\rQUANTITY : ";
      cin.getline(ip_STR, sizeof(ip_STR));
      } 
   temp.ProdQuan = atoi(ip_STR);
      
   cout << "UNIT PRICE (in pence)  : ";
   cin.getline(ip_STR, sizeof(ip_STR));
   while (!ip_STR[0])
      {
      // User isn't able to enter blank field
      cout << "UNIT PRICE (in pence)  : ";
      cin.getline(ip_STR, sizeof(ip_STR));
      } 
   temp.ProdCost = atoi(ip_STR);  

   Choose_supplier(SuppNumber, supp_FPTR_REF);
   temp.SuppNum = SuppNumber;
   
   stock_FPTR_REF.seekg(temp.ProdNum*sizeof(sProd),ios::beg);
   stock_FPTR_REF.write((char*) &temp, sizeof(sProd));
   }   

/***********************************************************
*                                                          *
*   FUNCTION: Browse_suppliers()                           *
*                                                          *
*   PURPOSE:  List and edit available suppliers.           *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
int Browse_suppliers(fstream& supp_FPTR_REF)
   { 
   // Pre - valid file pointer
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sSupp temp;
   char ip_STR[81];
   int choice;
   
   cout << "Current suppliers are: \n\n";
   
   for (int supp = 0; supp < MAX_SUPPLIER; supp++)
      {
      supp_FPTR_REF.seekg(supp*sizeof(sSupp), ios::beg);
      supp_FPTR_REF.read((char*) &temp, sizeof(sSupp));
      cout  << supp << " " << setw(15) << temp.SuppName;
      cout  << "  " << setw(15) << temp.SuppAddr << endl;
      }
   
   do {
      do 
         {
         cout << "\nType a number to edit entry, or "
              << supp << " to QUIT ";
         cin.getline(ip_STR, sizeof(ip_STR));
         } while ((!ip_STR[0]));

      choice = atoi(ip_STR);
      } while ((choice < 0) || (choice > MAX_SUPPLIER));
   
   if (choice < MAX_SUPPLIER)
      {
      do 
         {
         cout << "Enter new supplier name: ";
         cin.getline(temp.SuppName, sizeof(temp.SuppName));
         } while ((!temp.SuppName[0])); 

      do 
         {
         cout << "Enter new supplier SuppAddr: ";
         cin.getline(temp.SuppAddr, sizeof(temp.SuppAddr));
         } while ((!temp.SuppAddr[0])); 

      temp.SuppNum = choice;
      // Update disk copy
      supp_FPTR_REF.seekg(choice*sizeof(sSupp), ios::beg);
      supp_FPTR_REF.write((char*) &temp, sizeof(sSupp));

      Browse_suppliers(supp_FPTR_REF); //  Recursive call
      }
   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Choose_supplier()                            *
*                                                          *
*   PURPOSE:  Allow user to choose a supplier.             *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
int Choose_supplier(int& supp_num_REF,
                    fstream& supp_FPTR_REF)
   { 
   // Pre - valid file pointer
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sSupp temp;
   char ip_STR[81];
   int choice;
   
   cout << "Available suppliers are: \n\n";
   
   for (int supp = 0; supp < MAX_SUPPLIER; supp++)
      {
      supp_FPTR_REF.seekg(supp*sizeof(sSupp), ios::beg);
      supp_FPTR_REF.read((char*) &temp, sizeof(sSupp));
      cout  << supp << " " << setw(15) << temp.SuppName;
      cout  << "  " << temp.SuppAddr << endl;
      }
   
   do {
      do 
         {
         // User isn't able to enter blank field
         cout << "\nType a number to edit entry, or "
              << supp << " to QUIT ";
         cin.getline(ip_STR, sizeof(ip_STR));
         } while ((!ip_STR[0]));

      choice = atoi(ip_STR);
      } while ((choice < 0) || (choice > MAX_SUPPLIER));
  
   if (choice < MAX_SUPPLIER)
      {
      supp_num_REF = choice;
      }
   else
      {
      Browse_suppliers(supp_FPTR_REF);
      Choose_supplier(supp_num_REF, supp_FPTR_REF);
      }
      
   return 0;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Format_new_stock_file()                      *
*                                                          *
*   PURPOSE:  Fills empty data file with blank records.    *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Format_new_stock_file(fstream& stock_FPTR_REF)
   {
   // Pre - valid file pointer
   assert(stock_FPTR_REF.is_open());

   // Local variable
   sProd EmptyRecord = {0, "(BLANK)", 0, 0, 0};
   
   cout << "Formatting file stock.dat...\n";
      
   for (int stock_no = 1; stock_no <= MAX_STOCK; stock_no++)
      {
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.write((char*) &EmptyRecord, sizeof(sProd));
      }
       
   cout << "Format complete. \n" 
        << MAX_STOCK << " blank records created.\n";   
   }

/***********************************************************
*                                                          *
*   FUNCTION: Format_new_supplier_file()                   *
*                                                          *
*   PURPOSE:  Fills empty data file with blank records.    *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Format_new_supplier_file(fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointer
   assert(supp_FPTR_REF.is_open());

   // Local variable
   sSupp EmptyRecord = {0, "(EMPTY)", "(EMPTY)"};

   cout << "Formatting file 'supplier.dat'...\n";
      
   for (int supp = 0; supp < MAX_SUPPLIER; supp++)
      {
      supp_FPTR_REF.seekg(supp*sizeof(sSupp), ios::beg);
      supp_FPTR_REF.write((char*) &EmptyRecord, sizeof(sSupp));
      }
       
   cout << "Format complete. \n" 
        << MAX_SUPPLIER << " blank records created.\n";   
   }

/***********************************************************
*                                                          *
*   FUNCTION: Get_all_data()                               *
*                                                          *
*   PURPOSE:  Allows data entry into newly-created file.   *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Get_all_data(fstream& stock_FPTR_REF, 
                  fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sProd temp;
   int stock_no;

   cout << "\nPlease supply stock information, as follows:";
   
   Get_valid_ProdNum(stock_no);
    
   while (stock_no != -1) 
      {
      /* User doesn't want to quit 
         Check whether requested product entry exists */
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
 
      if (!temp.ProdNum) 
         {
         Fill_record(stock_FPTR_REF, supp_FPTR_REF, stock_no);
         }
      else
         {
         // Requested product entry DOES already exist   
         cout << "Prod number : " << temp.ProdNum << "\n";
         cout << "A product with the specified stock number ";
         cout << "already exists.\n";
         }               

      Get_valid_ProdNum(stock_no);
      }
   }

/***********************************************************
*                                                          *
*   FUNCTION: Menu()                                       *
*                                                          *
*   PURPOSE:  Displays user instructions                   *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
int Menu() 
   {
   // Local variables 
   int choice;
   char ip_STR[80];

   cout  << "\nMENU:\n"
         << "1  List all stock items.\n"
         << "2  List items with low stock level.\n"
         << "3  Update record for a particular item.\n"
         << "4  Insert record for a new item.\n"
         << "5  Delete record for an item.\n"
         << "6  Browse supplier details.\n"
         << "7  QUIT program.\n\n"
         << "? :";
         
   cin.getline(ip_STR, sizeof(ip_STR));    
   
   while ((!(sscanf(ip_STR, "%d", &choice) == 1)) 
         || (choice < 1) || (choice > 7))
      {
      cout << "\nOption must be in the range 1 - 7. "
           << "Please re-enter: ";
      cin.getline(ip_STR, sizeof(ip_STR));    
      }
   
   cout << endl;
   
   return choice;
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_all_records()                        *
*                                                          *
*   PURPOSE:  Lists all non-empty records                  *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_all_records(fstream& stock_FPTR_REF,
                         fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sProd temp;
   int number_of_stock_records = 0;
   long total_value = 0;  // Total stock value (in pence)

   for (int stock_no =1; stock_no <= MAX_STOCK; stock_no++) 
      {
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
      
      if (temp.ProdNum)
         {
         number_of_stock_records++;
            
         if (number_of_stock_records == 1)
            {
            Display_record_header();
            }
         total_value += (long) temp.ProdQuan * temp.ProdCost;   
         Display_record(temp, supp_FPTR_REF);
         }
      }

   if (number_of_stock_records == 0)
      {        
      cout << "The stock inventory is empty. \n";
      }
   else
      {
      Display_record_footer();
      
      cout  << "\nThe total value of the inventory is ";
      Display_money(total_value);
      cout << endl;
      }        
   
   Type_enter_to_continue();
   } 

/***********************************************************
*                                                          *
*   FUNCTION: Get_valid_ProdNum()                          *
*                                                          *
*   PURPOSE:  Prompt user for a valid stock number.        *
*                                                          *
*   PRE:      None.                                        *
*   POST:     stock_no_REF is valid.                       *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Get_valid_ProdNum(int& stock_no_REF)
   {
   // Local variable 
   char ip_STR[81];

   cout  << "\nType a stock number (1-" 
         << MAX_STOCK << "), or -1 to end input): ";

   cin.getline(ip_STR, sizeof(ip_STR));    
   
   while ((!(sscanf(ip_STR, "%d", &stock_no_REF) == 1)) 
            || (stock_no_REF < -1) 
            || (stock_no_REF > MAX_STOCK)
            || (stock_no_REF == 0)) 
      {
      cout  << "\nStock number must be in range 1 - " 
            << MAX_STOCK << ". Please re-enter: ";

      cin.getline(ip_STR, sizeof(ip_STR));    
      }

   // Post - stock number is valid
   assert ((stock_no_REF > -2)
          && (stock_no_REF <= MAX_STOCK));
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_money()                              *
*                                                          *
*   PURPOSE:  Display (pounds.pence) amount in pence.      *
*                                                          *
*   PRE:      None.                                        *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_money(const long MONEY)
   {
   cout  << "" << MONEY / ((long) 100)
         << ".";
             
   int pence = (int)(MONEY % ((long) 100));      
   //Make sure "pence" output is displayed correctly
   //E.g. 123.07 cf. 123.7 (which would be ambiguous)
   if (pence < 10)
      { 
      //Display initial 0
      cout << "0" << pence << endl;
      }
   else
      {        
      cout << pence;
      }
   }

/***********************************************************
*                                                          *
*   FUNCTION: Display_low_stock_records()                  *
*                                                          *
*   PURPOSE:  Lists all products with low stock levels.    *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Display_low_stock_records(fstream& stock_FPTR_REF,
                               fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sProd temp;
   int number_of_stock_records = 0;
   int number_of_low_stock_records = 0;
   
   for (int stock_no = 1; stock_no <= MAX_STOCK; stock_no++)
      {
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
      
      if (temp.ProdNum)
         {
         number_of_stock_records++;

         if (temp.ProdQuan < LOW_STOCK_LEVEL)
            {   
            number_of_low_stock_records++;
            
            if (number_of_low_stock_records == 1)
               {
               cout << "Stock of these products is low: \n\n";
               Display_record_header();
               }
            Display_record(temp, supp_FPTR_REF);
            }
         }
      }

    if (number_of_stock_records == 0)
      {        
      cout << "The stock inventory is empty. \n";
      }
    else
      {  
      if (number_of_low_stock_records == 0)
         {        
         cout << "All stock levels are >= "
              << LOW_STOCK_LEVEL << " units. \n";
         }
      else
         {
         Display_record_footer();
         }   
      }
      
   Type_enter_to_continue();
   }

/***********************************************************
*                                                          *
*   FUNCTION: Update_record()                              *
*                                                          *
*   PURPOSE:  Updates contents of a specified record.      *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Update_record(fstream& stock_FPTR_REF, 
                   fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sProd temp;
   int stock_no;
   char ip_STR[30];
   
   Get_valid_ProdNum(stock_no);

   if (stock_no != -1)  // -1 => quit
      {
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
   
      if (temp.ProdNum) 
         {
         Display_record_header();
         Display_record(temp, supp_FPTR_REF);
     
         cout << "New NAME (or <ENTER> to leave unchanged)     : ";
         cin.getline(ip_STR, sizeof(ip_STR));    
      
         if (*ip_STR)
            {
            // input was NOT blank, therefore update entry 
            strcpy(temp.ProdName, ip_STR);
            }
      
         cout << "New QUANTITY (or <ENTER> to leave unchanged) : ";
         cin.getline(ip_STR, sizeof(ip_STR));
      
         if (*ip_STR)
            {
            // input was NOT blank, therefore update entry 
            temp.ProdQuan = atoi(ip_STR);
            } 
                                   
         cout << "New PRICE (or <ENTER> to leave unchanged)    : ";
         cin.getline(ip_STR, sizeof(ip_STR));
      
         if (*ip_STR)
            {
            // input was NOT blank, therefore update entry 
            temp.ProdCost = atoi(ip_STR);   
            } 
         
         stock_FPTR_REF.seekg(temp.ProdNum*sizeof(sProd),ios::beg);
         stock_FPTR_REF.write((char*) &temp, sizeof(sProd));

         Display_record_header();
         Display_record(temp, supp_FPTR_REF);
         }
      else
         {
         cout << "Cannot update. The record is empty.\n"; 
         }                     
      }               
      
   Type_enter_to_continue();
   }

/***********************************************************
*                                                          *
*   FUNCTION: Insert_record()                              *
*                                                          *
*   PURPOSE:  Allows a new record to be inserted in file.  *
*                                                          *
*   PRE:      Valid file pointers.                         *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Insert_record(fstream& stock_FPTR_REF, 
                   fstream& supp_FPTR_REF)
   {
   // Pre - valid file pointers
   assert(stock_FPTR_REF.is_open());
   assert(supp_FPTR_REF.is_open());

   // Local variables
   sProd temp;
   int stock_no;
   
   Get_valid_ProdNum(stock_no);
    
   stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
   stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
   
   if (!temp.ProdNum) 
      { 
      Fill_record(stock_FPTR_REF, supp_FPTR_REF, stock_no);
      }
   else
      {
      cout << "A product with the specified stock number ";
      cout << "already exists.\n";
      }
   
   Type_enter_to_continue();
   }

/***********************************************************
*                                                          *
*   FUNCTION: Delete_record()                              *
*                                                          *
*   PURPOSE:  Allows deletion of a specified record.       *
*                                                          *
*   PRE:      Valid file pointer.                          *
*   POST:     None.                                        *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Delete_record(fstream& stock_FPTR_REF)
   {
   // Pre - valid file pointer
   assert(stock_FPTR_REF.is_open());

   // Local variables
   sProd EmptyRecord = {0, "(EMPTY)", 0, 0}, temp;
   int stock_no;

   Get_valid_ProdNum(stock_no);

   if (stock_no != -1)
      {
      stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
      stock_FPTR_REF.read((char*) &temp, sizeof(sProd));
   
      if (temp.ProdNum) 
         {
         stock_FPTR_REF.seekg(stock_no*sizeof(sProd), ios::beg);
         stock_FPTR_REF.write((char*) &EmptyRecord, sizeof(sProd));
         cout << "Record deleted.\n";
         }
      else
         {
         cout << "Cannot delete. The record is empty.\n";
         }
      }
      
   Type_enter_to_continue();
   }

/***********************************************************
*                                                          *
*   FUNCTION: Open_files()                                 *
*                                                          *
*   PURPOSE:  Open database.                               *
*                                                          *
*   PRE:      None.                                        *
*   POST:     Got valid file pointers.                     *
*                                                          *
*   RETURNS:  void.                                        *
*                                                          *
***********************************************************/
void Open_files(fstream& stock_FPTR_REF,
                fstream& supp_FPTR_REF)
   {
   // Local variable 
   char ip_STR[81];

   // Note the 'nocreate' option is not provided in
   // the draft C++ Standard, which makes this code
   // rather clumsy... [MJP 12 May, 1997]

   // Try to find an existing data file "supplier.dat" - 
   // if it doesn't exist (or can't find it), 
   // offer user option of creating one.
   supp_FPTR_REF.open("supplier.dat", ios::in);
   
   if (supp_FPTR_REF.fail())
      {   
      // Can't find an existing supplier file 
      cout << "\nCan't find the data file 'supplier.dat'\n";
   
      // Ask user if she wants to create a new file 
      cout << "\nDo you want to create a new file (Y or N): ";
      cin.getline(ip_STR, sizeof(ip_STR));    
      
      while ((toupper(ip_STR[0]) != 'Y') && 
             (toupper(ip_STR[0]) != 'N')) 
         {
         cout << "Invalid response. Enter Y or N: ";
         cin.getline(ip_STR, sizeof(ip_STR));    
         }
      
      if (toupper(ip_STR[0]) == 'Y') 
         {
         supp_FPTR_REF.clear();
         supp_FPTR_REF.open("supplier.dat", ios::out
                                | ios::in | ios::binary);
   
         if (supp_FPTR_REF.fail())
            {
            cout << "\nFATAL ERROR:\n";
            cout << "File 'supplier.dat' could not be created.\n";
            cout << "Is the disk full?\n";
            exit(1);
            }
          
         Format_new_supplier_file(supp_FPTR_REF);
         Browse_suppliers(supp_FPTR_REF);
         }
     else
         {
         // No supplier file exists, and user does not 
         // want to create a new one: no option but to QUIT 
         cout << "Please restore backup of 'supplier.dat'\n";
         cout << "Then re-run this program.\n";
         exit(1);      
         } 
      }  

   // Close the file
   supp_FPTR_REF.close();
   // Re-open the file in correct mode 
   supp_FPTR_REF.open("supplier.dat", ios::out | ios::in
                             | ios::binary );

   if (supp_FPTR_REF.fail())
      {
      cerr << "\nFATAL ERROR:\n";
      cerr << "File 'supplier.dat' could not be opened.\n";
      exit(1);
      }

   // Try to find an existing data file "stock.dat" - 
   // if it doesn't exist (or can't find it), 
   // offer user option of creating one 
   stock_FPTR_REF.open("stock.dat", ios::in);   
   
   if (stock_FPTR_REF.fail())
      {   
      // Can't find an existing stock file 
      cout << "\nCan't find the data file 'stock.dat'\n";
   
      // Ask user if s/he wants to create a new file 
      cout << "\nDo you want to create a new file (Y or N): ";
      cin.getline(ip_STR, sizeof(ip_STR));    
      
      while ((toupper(ip_STR[0]) != 'Y') && 
             (toupper(ip_STR[0]) != 'N')) 
         {
         cout << "Invalid response. Enter Y or N: ";
         cin.getline(ip_STR, sizeof(ip_STR));    
         }
      
      if (toupper(ip_STR[0]) == 'Y') 
         {
         stock_FPTR_REF.clear();
         stock_FPTR_REF.open("stock.dat", ios::out
                                        | ios::in | ios::binary);   
   
         if (stock_FPTR_REF.fail())
            {
            cerr << "\nFATAL ERROR:\n";
            cerr << "File 'stock.dat' could not be created.\n";
            cerr << "Is the disk full?\n";
            exit(1);
            }
          
         Format_new_stock_file(stock_FPTR_REF);
         Get_all_data(stock_FPTR_REF, supp_FPTR_REF);
         }
     else
         {
         // No stock file exists, and user does not 
         // want to create a new one: no option but to QUIT 
         cout << "Please restore backup of 'stock.dat'\n";
         cout << "Then re-run this program.\n";
         exit(1);      
         } 
      }  

   // Close the file
   stock_FPTR_REF.close();
   // Open the file in correct mode >>>
   stock_FPTR_REF.open("stock.dat", ios::out | ios::in
                             | ios::binary );

   if (stock_FPTR_REF.fail())
      {
      cerr << "\nFATAL ERROR:\n";
      cerr << "File 'stock.dat' could not be opened.\n";
      exit(1);
      }

   // Post - valid file pointers
   assert(stock_FPTR_REF.good());
   assert(supp_FPTR_REF.good());
   }

/***********************************************************
*                   *** END OF PROGRAM ***                 *
***********************************************************/
