// A Fragment of a Mailing List Program
// Demonstrating classes and operator overloading 
// Source is: 24_02.CPP

#include <iostream.h>
#include <stdlib.h>

// Maximum size of the "database"
const int MAX_LIST_SIZE = 3;

// The customer class
class cCustomer 
   {
   public:
      friend istream &operator>> (istream&, cCustomer&);
      friend ostream &operator<< (ostream&, const cCustomer&);

   private:
      char _Last_name[81];  // Customer's last name           
      char _First_name[81]; // Customer's first name          
      char _Title[81];      // Customer's title               
      char _Address1[81];   // 1st line of customer's addr 
      char _Address2[81];   // 2nd line of customer's addr     
      char _Address3[81];   // 3rd line of customer's addr      
      char _PostCode[81];   // Customer's postcode            
   };

/***********************************************************
*                                                          *
*   OPERATOR FUNCTION: >>                                  *
*                                                          *
*   PURPOSE: Overload ">>" operator.                       *
*                                                          *
***********************************************************/
istream &operator>> (istream& input, cCustomer& Cust)
   {
   cout << "Enter customer's last name         : ";  
   input.getline(Cust._Last_name, 80);
      
   cout << "Customer's first name              : ";
   input.getline(Cust._First_name, 80); 
           
   cout << "Customer's title (e.g. Ms Mr)      : ";
   input.getline(Cust._Title, 80);
 
   cout << "First line of customer's address   : ";
   input.getline(Cust._Address1, 80);

   cout << "Second line of customer's address  : ";
   input.getline(Cust._Address2, 80); 

   cout << "Third line of customer's address   : ";
   input.getline(Cust._Address3, 80);
       
   cout << "Customer's postcode                : ";
   input.getline(Cust._PostCode, 80);
   cout << endl;

   return input;
   } 

/***********************************************************
*                                                          *
*   OPERATOR FUNCTION: <<                                  *
*                                                          *
*   PURPOSE: Overload "<<" operator.                       *
*                                                          *
***********************************************************/
ostream &operator<< (ostream& output, const cCustomer& CUST)
   {
   output << CUST._Title << " "; 
   output << CUST._First_name << " "; 
   output << CUST._Last_name << endl; 
   output << CUST._Address1 << endl;
   output << CUST._Address2 << endl;
   output << CUST._Address3 << endl;
   output << CUST._PostCode << "\n\n";

   return output;
   } 


// Ordinary function prototype
int menu(int&);

int main() 
   { 
   cCustomer cust[MAX_LIST_SIZE]; // Array of objects
   int list_size = 0;  // Size of the list

   int choice;

   cout << "OBJECT-BASED MAILING LIST\n";
   cout << "-------------------------\n\n";
 
   do {
      if (menu(choice))
         {
         cerr << "FATAL ERROR in function menu()";
         exit(1);
         }
      else
         {
         switch(choice) 
            {
            case 1: // Enter new name      
               if (list_size <= (MAX_LIST_SIZE-1))
                  {
                  cin >> cust[list_size];  // Note !!
                  list_size++;
                  }
               else
                  {
                  cout << "Sorry - the database is full\n\n";
                  }
               break;            

            case 2: // Display (all) names
               if (list_size)
                  {
                  for (int  c = 0; c < list_size; c++)
                     {
                     cout << cust[c];  // Note !!
                     }
                  }
               else
                  {
                  cout << "Sorry - database is empty\n\n";
                  }
            }
         }   
      } while(choice != 3);

   cout << "Finished.";

   return 0;   
   }

/***********************************************************
*                                                          *
*   FUNCTION: menu                                         *
*                                                          *
*   PURPOSE: Offer user options and return a selection.    *
*                                                          *
***********************************************************/
int menu(int& choice)
   {
   char text_str[81];

   cout << " 1. Enter New Data\n";
   cout << " 2. Display All Data\n";
   cout << " 3. Quit\n\n";

   do 
      {
      cout << "? : ";
      cin.getline(text_str, 80);
      choice = text_str[0] - '0'; 
      } while((choice < 1) || (choice > 3));
   
   return 0; 
   }
