// A new dynamic array data type
// Source is: 25_10.CPP

#include <iostream.h>
#include <stdlib.h>

class cIntArray
   {
   public: 
      // Construction / destruction
      cIntArray(const int = 1);    // Constructor  
      cIntArray(const cIntArray&); // Copy constructor
      ~cIntArray();                // Destructor
      
      // General member functions
      int& Access(const int& i);
      int  Get_Size() const;
      
   private:
      int* _Start_PTR;
      int  _Size;   
   };

cIntArray::cIntArray(const int SIZE)
   {
   cout << "   Using normal constructor\n";
   _Size = SIZE;

   if (_Size < 1)
      { 
      cerr << "\n*** Array size " << _Size << " cant be created. ***\n";
      exit(1);
      }
   else
      {
      _Start_PTR = new int[_Size];
   
      if ( _Start_PTR == 0)
         {
         cerr << "\n*** Unable to allocate memory - aborting ***\n";
         exit(1);
         }
      else
         {
         // Initialise new array
         for (int i = 0; i < _Size; i++)
            {
            _Start_PTR[i] = 0;
            }
         }   
      }
   }   

// Copy constructor
cIntArray::cIntArray(const cIntArray& A)
   {
   cout << "   Using copy constructor\n"; 
   
   _Size = A._Size;
   
   // Memory is allocated for the copy
   _Start_PTR = new int[A._Size];  
   
   if(!_Start_PTR)
      {
      cout << "ERROR!";
      exit(1);
      } 
      
   for (int i = 0; i < A._Size; i++)
      {
       _Start_PTR[i] = A._Start_PTR[i];
      }
   }
      
// Destructor      
cIntArray::~cIntArray()
   {
   cout << "   Using destructor\n";
   delete [] _Start_PTR;   
   _Start_PTR = 0;
   }
 
// Access 'two-way' member function 
int& cIntArray::Access(const int& I)
   {
   if ((I < 0) || (I >=  _Size))
      {
      cerr << "\n*** Array out of bounds error ***\n";
      exit(1);
      }
                                  
   return (_Start_PTR[I]);
   }

// Get_Size member function   
int cIntArray::Get_Size() const
   {
   return (_Size);
   }

// Prototype for ordinary function 
// (for illustrative purposes only)
void Display_Array(cIntArray);

int main()
   {
   // Create a couple of arrays
   cout << "About to create two new arrays\n";
   cIntArray one(4), two(3);
  
   one.Access(1) = 2; // Set values of some elements
   two.Access(2) = 999;  
   
   // Create new array 
   cout << "About to create new (copy) array\n";
   cIntArray three = one;
   
   cout << "About to call Display_Array()\n";
   Display_Array(one); 
   cout << "About to call Display_Array()\n";
   Display_Array(three);

   cout << "About to leave main()\n";
   return 0;
   }   

void Display_Array(cIntArray Array)
   {
   cout << "Just entered Display_Array()\n";
   
   for (int i = 0; i < Array.Get_Size(); i++)
      {
      cout  << "   Element " << i << " = " 
            << Array.Access(i) << endl;
      }
   
   cout << "Just leaving Display_Array()\n";
   }
