// Copying simple arrays and class-based arrays
// Compares the speed of each technique. 
// Source is: 25_12.CPP
   
// Fast version with inline Access function & assertions

#define NDEBUG // Switch off assertions

#include <assert.h>
#include <iostream.h>   
#include <time.h>
#include <stdlib.h>

class cIntArray
   {
   public:  
      // Construction / destruction
      cIntArray(const int = 1);
      cIntArray(const cIntArray&);
      ~cIntArray();
      
      // General member functions
      inline int& Access(const int&);
             int  Get_Size();
             
   private:
      int* _Start_PTR;
      int  _Size;   
   };

cIntArray::cIntArray(const int SIZE)
   {
   _Size = SIZE;
     
   if (_Size < 1)
      {
      cerr << "\n*** Array size " << _Size << " cant be created. ***\n";
      exit(1);
      }
   else
      {
      _Start_PTR = new int[_Size];
   
      if (_Start_PTR == 0)
         {
         cerr << "\n*** Unable to allocate memory - aborting ***\n";
         exit(1);
         }
      else
         {
         // Initialise new array
         for (int i = 0; i < _Size; i++)
            {
            _Start_PTR[i] = 0;
            }
         }   
      }
   }   
      
// Copy constructor
cIntArray::cIntArray(const cIntArray& A)
   {
   _Size = A._Size;
   
   // Memory is allocated for the copy
   _Start_PTR = new int[A._Size];  
   
   if(! _Start_PTR)
      {
      cout << "ERROR!";
      exit(1);
      } 
      
   for (int i = 0; i < A._Size; i++)
      {
       _Start_PTR[i] = A._Start_PTR[i];
      }
   }

cIntArray::~cIntArray()
   {
   delete [] _Start_PTR;   
   _Start_PTR = 0;
   }

int& cIntArray::Access(const int& I)
   {
   // Precond (bounds checking)
   assert ((I >= 0) && (I <  _Size));
                                  
   return (_Start_PTR[I]);
   }
   
int cIntArray::Get_Size()
   {
   return (_Size);
   }

int main()
   {
   const int   DIM = 1000;
   const long  NUM_PASSES = 10000;

   int      Orig[DIM] = {0};
   int      Copy[DIM]; 
   cIntArray Orig2(DIM);  // IDT Arrays
   cIntArray Copy2(DIM);
   
   double   elapsed_time;
   time_t   start_time, stop_time; // time_t is defined in time.h

   cout << "Starting first calculation - please wait...\n";
    
   // METHOD ONE
   start_time = time(NULL);
   
   for (long pass = 0; pass <= NUM_PASSES; pass++)
      { 
      for (int i = 0; i < DIM; i++)
         {
         Copy[i] = Orig[i];
         }
      }
             
   stop_time = time(NULL);   
   elapsed_time = difftime(stop_time, start_time);
   cout << "Using normal arrays, time (in seconds) : " 
        << elapsed_time << endl; 

   cout << "Starting second calculation - please wait...\n";
    
   // METHOD TWO
   start_time = time(NULL);
   
   for (pass = 0; pass <= NUM_PASSES; pass++)
      { 
      for (int i = 0; i < DIM; i++)
         {
         Copy2.Access(i) = Orig2.Access(i);
         }
      }
             
   stop_time = time(NULL);   
   elapsed_time = difftime(stop_time, start_time);
   cout << "Using class-based arrays, time (in seconds) : " 
        << elapsed_time << endl; 
    
   return 0;    
   }
