// A new dynamic array data type this time with exceptions
// Note also the appropriate use of a private member function
// Source is: 25_15.CPP

#include <iostream.h>
#include <stdlib.h>

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cIntArray                                      *
*                                                          *
*   EXCEPT: Throws 1 (int) on error                        *
*                                                          *
************************************************************
***********************************************************/
class cIntArray
   {
   public:
           cIntArray(const int = 1);    // Constructor  
           cIntArray(const cIntArray&); // Copy constructor
           ~cIntArray();                // Destructor
           
      int& Access(const int&);
      int  Get_Size();

   private:
      void New_Memory();  // Group memory handling here

      int* _Start_PTR;
      int  _Size;   
   };

cIntArray::cIntArray(const int SIZE)
   {
   if (SIZE < 1)
      { 
      cerr << "\n*** Array size " << SIZE << " cant be created. ***\n";
      throw 1;
      }
   else
      {
      _Size = SIZE;
      
      New_Memory();  // Grab memory from free store
      
      // Initialise new array
      for (int i = 0; i < _Size; i++)
         {
         _Start_PTR[i] = 0;
         }
      }
   }   

// Copy constructor
cIntArray::cIntArray(const cIntArray& A)
   {
   _Size = A._Size;
   
   // Memory is allocated for the copy
   New_Memory();
   
   // Copy 'old' data to 'new' array
   for (int i = 0; i < _Size; i++)
      {
      _Start_PTR[i] = A._Start_PTR[i];
      }
   }
      
cIntArray::~cIntArray()
   {
   delete []  _Start_PTR;
   _Start_PTR = 0;   
   }

int& cIntArray::Access(const int& I)
   {
   if ((I < 0) || (I >=  _Size))
      {
      throw 1;
      }
                                  
   return ( _Start_PTR[I]);
   }
   
int cIntArray::Get_Size()
   {
   return ( _Size);
   }

void cIntArray::New_Memory()
   {
   _Start_PTR = new int[_Size];  
   
   if(!_Start_PTR)
      {
      throw 1;  // Out of memory...
      } 
   }

int main()
   {
   try 
      { 
      // Create an array
      cIntArray one(100);
  
      one.Access(1) = 1;   // Set values of some elements
      one.Access(99) = 99;

      cout << "Position 1   : " << one.Access(1) << endl;
      cout << "Position 99  : " << one.Access(99) << endl;
      cout << "Position 100 : " << one.Access(100) << endl;
      cout << "Position 7   : " << one.Access(7) << endl;
      }

   catch (...)
      {
      cout << "\n*** An exception has occurred ***";
      return 1;
      }

   return 0;
   }   
