//
// The file demonstrates an extended version of the 
// cIntArray class in file 25_15.cpp
// 
// cDynIntArray is 'more dynamic': it automatically
// grows in size (within the limits of available free
// store memory) as new elements are required.
//
// Source is: 25_15a.CPP
//

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cDynIntArray                                   *
*                                                          *
*   EXCEPT: Throws 1 (int) on error                        *
*                                                          *
************************************************************
***********************************************************/
class cDynIntArray
   {
   public: 
      // Construction / destruction
      cDynIntArray(const int = 0);       // Constructor  
      cDynIntArray(const cDynIntArray&); // Copy constructor
      ~cDynIntArray();                   // Destructor
 
      // General member functions          
      int& Access(const int&);
      int  Get_Size();
      void Set_Size(const int);   // New member function

   protected:
      void New_Memory();  // Group memory handling here

   private:
      int* _pStart;
      int  _Size;   
   };

cDynIntArray::cDynIntArray(const int SIZE)
   {
   // Can't create arrays of size < 0...
   assert (SIZE >= 0);

   cout << "*** Initial size of array is " << SIZE << " ***\n";

   _Size = SIZE;

   if (_Size == 0)
      { 
      return;  // Don't need any memory...
      }
   else
      {
      New_Memory();  // Grab memory from free store
      
      // Initialise new array
      for (int i = 0; i < _Size; i++)
         {
         _pStart[i] = 0;
         }
      }
   }   

// Copy constructor
cDynIntArray::cDynIntArray(const cDynIntArray& A)
   {
   _Size = A._Size;
   
   // Don't do anything else if orig array is size 0
   if (_Size > 0)
      {
      // Memory is allocated for the copy
      New_Memory();
   
      // Copy 'old' data to 'new' array
      for (int i = 0; i < _Size; i++)
         {
         _pStart[i] = A._pStart[i];
         }
      }
   }
      
cDynIntArray::~cDynIntArray()
   {
   if (_Size > 0)
      {
      delete []  _pStart;
      _pStart = 0;   
      }
   }

int& cDynIntArray::Access(const int& I)
   {
   if (I < 0)
      {
      cerr << "Trying to create array with size < 0\n";
      throw 1;
      }
   
   // If 'out of bounds' increase size of array
   if (I >= _Size)
      {
      cout << "*** Increasing array size to " << I+1 
           << " ***" << endl;
      Set_Size(I + 1);
      }

   return (_pStart[I]);
   }
   
int cDynIntArray::Get_Size()
   {
   return _Size;
   }

void cDynIntArray::Set_Size(const int NEW_SIZE)
   {
   // Don't do anything if new size == old size
   if (NEW_SIZE == _Size)
      {
      return;
      }

   int* Tmp_PTR = new int[NEW_SIZE];

   if(!Tmp_PTR)
      {
      cerr << "Out of memory.\n";
      throw 1;  // Out of memory...
      } 

   // Find smaller of new size or old size
   int Min_Size = _Size < NEW_SIZE ? _Size : NEW_SIZE;

   // Copy approp. elements from old to new
   // areas of memory
   for (int i = 0; i < Min_Size; i++)
      {
      // If the new size is smaller than the old
      // size then data will be lost - without warning...
      Tmp_PTR[i] = _pStart[i];
      }

   // Initialise new elements (if any)
   for (i = _Size; i < NEW_SIZE; i++)
      {
      Tmp_PTR[i] = 0;
      }

   // If there was an old array, delete it
   if (_Size > 0)
      {
      delete [] _pStart;
      }

   _Size = NEW_SIZE;
   _pStart = Tmp_PTR;
   }

void cDynIntArray::New_Memory()
   {
   _pStart = new int[_Size];  
   
   if(!_pStart)
      {
      cerr << "Out of memory..\n";
      throw 1;  // Out of memory...
      } 
   }

int main()
   {
   try 
      { 
      // Create an array (initial size 0)
      cDynIntArray One;
  
      One.Access(1) = 1;   // Set values of some elements
      cout << "Size of array is now " << One.Get_Size() << endl;

      // This will automatically enlarge the array
      One.Access(99) = 99;
      cout << "Size of array is now " << One.Get_Size() << endl;

      cout << "Position 1   : " << One.Access(1) << endl;
      cout << "Position 99  : " << One.Access(99) << endl;
      cout << "Position 100 : " << One.Access(100) << endl;

      One.Access(7) = 999;
      cout << "Position 7   : " << One.Access(7) << endl;

      // Array grows in size as required, and can be
      // manually reduced in size if necessary
      cout << "*** Manually reducing size of the array ***\n";
      One.Set_Size(8);
      cout << "Size of array is now " << One.Get_Size() << endl;
      cout << "Position 7   : " << One.Access(7) << endl;
      }

   catch (...)
      {
      cout << "\n*** An exception has occurred ***";
      return 1;
      }

   return 0;
   }   
