//
// A class example using a copy constructor
// and overloaded assignment operator
// Adapted from 25_09.cpp

// Source is: ASS_COPY.CPP

// NOTE how checks for self assignment are carried 
// in the copy constructor and overloaded assignment
// operator - see main for an explanation of why 
// such checks should be included in professional code.

// MJP 
// 14 March, 1997

#include <assert.h>
#include <iostream.h>
#include <stdlib.h>

class cDynamic
   {
   public:
      // General member functions    
      int  Show_Num() const;

      // Overloaded =
      cDynamic operator=(const cDynamic&);
          void operator=(const int);

      // Construction / destruction
      cDynamic(int = 0);         // Default constructor
      cDynamic(const cDynamic&); // Copy constructor
      ~cDynamic();               // Destructor
      
   protected:
      void New_Memory();

   private:
      int* _pNum;
   };   

cDynamic::cDynamic(int Number)
   {
   _pNum = new int;
   
   New_Memory();

   *_pNum = Number;
   }
                     
cDynamic::cDynamic(const cDynamic& rDyn)
   {
   // Check for self-assignment
   assert(this != &rDyn); 
   
   New_Memory();
   
   *_pNum = *(rDyn._pNum);
   }

cDynamic::~cDynamic()
   {
   cout << "*** Destructing a cDynamic object ***\n";
   delete _pNum;
   _pNum = 0;
   }


void cDynamic::New_Memory()
   {
   _pNum = new int;
   
   if (_pNum == NULL)
      {
      cerr << "Unable to allocate memory - aborting program\n";
      throw 1; // Throw exception (int here - could be anything)
      }
   }

cDynamic cDynamic::operator=(const cDynamic& rDyn)
   {
   // Check for self-assignment
   assert(this != &rDyn);

   *_pNum = *rDyn._pNum;
   
   return *_pNum;
   }

void cDynamic::operator=(const int I)
   {
   *_pNum = I;
   }

int cDynamic::Show_Num() const
   {
   return *_pNum;
   }

int main()
   { 
   // --------------------------------------------------
   // 
   // Note that because the following code (and similar) 
   // is syntactically valid:
   //
   int x = x;
   x = x;
   // ... you should include checks to prevent problems
   // arising in your code if instances of your classes
   // are used in this manner.
   //
   // Please refer to the copy constructor and 
   // overloaded assignment operator code for details. 
   //
   // --------------------------------------------------

   cout << "NOTE: This program will generate assertion failures..." << endl;
     
   try
      {  
   	  cDynamic test1;

      // Attempt self-assignment
      test1 = test1;  // Assertion failure...
 
      cDynamic test2 = test2;  // Assertion failure...
      }
   
   catch (...)
      {
      cout << "An error has occurred...";
      }
         
   return 0;
   }

