// A new generic dynamic array data type
// Source is: 28_05.CPP

#include <iostream.h>
#include <stdlib.h>
#include <assert.h>

template <typename TYPE, const int SIZE>
class cGenericArray
   {
   public:
      // Construction and destruction
      cGenericArray();                     // Default constructor 
      cGenericArray(const cGenericArray&); // Copy constructor
      ~cGenericArray();                    // Destructor
      
      // General member functions
      TYPE& operator[](const int); // Overload [] operator
      int   Get_Size() const;

   private:
      void  New_Memory(); // Place common memory handling here

      TYPE* _Start_PTR;
      int   _Size;
   };

template <typename TYPE, const int SIZE>
cGenericArray<TYPE, SIZE>::cGenericArray()
   {
   assert(SIZE > 0); // Precond
   
   _Size = SIZE;

   New_Memory();  // Grab memory from free store
   }   

// Copy constructor
template <typename TYPE, const int SIZE>
cGenericArray<TYPE, SIZE>::cGenericArray(const cGenericArray<TYPE, SIZE>& A)
   {
   _Size = A._Size;
   
   New_Memory(); // Memory is allocated for the copy
      
   for (int i = 0; i < _Size; i++)
      {
      _Start_PTR[i] = A._Start_PTR[i];
      }
   }    
      
template <typename TYPE, const int SIZE>
cGenericArray<TYPE, SIZE>::~cGenericArray()
   {
   delete [] _Start_PTR;
   _Start_PTR = 0;   
   }

template <typename TYPE, const int SIZE>
TYPE& cGenericArray<TYPE, SIZE>::operator[](const int i)
   {
   assert((i >=0) && (i <_Size)); // Precond

   return (_Start_PTR[i]);
   }
   
template <typename TYPE, const int SIZE>
int cGenericArray<TYPE, SIZE>::Get_Size() const
   {
   return (_Size);
   }

template <typename TYPE, const int SIZE>
void cGenericArray<TYPE, SIZE>::New_Memory()
   {
   _Start_PTR = new TYPE[_Size];  
   
   if(!_Start_PTR)
      {
      throw 1;  // Out of memory...
      } 
   }

int main()
   {
   try
      {
      // Create a couple of integer arrays
      cout << "About to create two new integer arrays\n";
      cGenericArray<int,4> one, two;
  
      one[3] = 0;    // Set values of some elements
      two[2] = 999;

      cout << one[3] << endl;
      cout << two[2] << endl;

      // Create new integer array 
      cout << "About to create new (copy) integer array\n";
      cGenericArray<int,4> three = one;
      cout << three[3] << endl;

      // Create a 2-D array
      cout << "About to create 2-D array\n";
      cGenericArray<cGenericArray<int,3>,3> four;

      four[1][1] = 3;
      cout << four[1][1];
      }

   catch (...)
      {
      cout << "An error has occurred..." << endl;
      return 1;
      }

   return 0;
   }
