/***********************************************************
*                                                          *
* A Simple (object-based) Mailing List Program             *
*                                                          *
* Source is:     31_03a.CPP                                *
* Created:       11 Dec, 1994                              *
* Last edited:   12 Mar, 1997                              *
*                                                          *
***********************************************************/

// Slightly 'purer' O-O version than that listed in the book
// No ordinary functions...

#include <assert.h>      
#include <fstream.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>

const int STRG_SZ = 81;   // Size of character arrays...

/***********************************************************
************************************************************
*                                                          *
*   CLASS:      cCustomer                                  *
*                                                          *
************************************************************
***********************************************************/
class cCustomer 
   {
   public:
      int Compare_Name(char*) const;
      int Compare_Area(char*) const;

      friend istream &operator>> (istream&, cCustomer&);
      friend ostream &operator<< (ostream&, const cCustomer&); 

   private:
      char _lastname[STRG_SZ];      // Last name
      char _firstname[STRG_SZ];     // First name
      char _title[STRG_SZ];         // Title
      char _address1[STRG_SZ];      // 1st line of address
      char _address2[STRG_SZ];      // 2nd line of address
      char _address3[STRG_SZ];      // 3rd line of address
      char _postcode[STRG_SZ];      // Postcode
   };


/***********************************************************
*                                                          *
*   O FUNCTION: >>                                         *
*                                                          *
*   CLASS:      cCustomer                                  *
*                                                          *
*   PURPOSE:    Overload ">>" operator.                    *
*                                                          *
*   RETURNS:    istream&                                   *
*                                                          *
***********************************************************/
istream& operator>> (istream& input,
                     cCustomer& cust_REF)
   {
   cout << "Enter last name (ENTER to quit)  : ";
   input.getline(cust_REF._lastname, STRG_SZ);
      
   if(!(cust_REF._lastname[0] == '\0'))  
      { // User didn't just hit <ENTER> - so continue
      cout << "First name                       : ";
      input.getline(cust_REF._firstname, STRG_SZ);
           
      cout << "Title (e.g. Ms Mr)               : ";
      input.getline(cust_REF._title, STRG_SZ);
 
      cout << "First line of address            : ";
      input.getline(cust_REF._address1, STRG_SZ);

      cout << "Second line of address           : ";
      input.getline(cust_REF._address2, STRG_SZ);

      cout << "Third line of address            : ";
      input.getline(cust_REF._address3, STRG_SZ);
      
      cout << "Post (zip) code                  : ";
      input.getline(cust_REF._postcode, STRG_SZ);
      }
   
   return input;
   } 

/***********************************************************
*                                                          *
*   O FUNCTION: <<                                         *
*                                                          *
*   CLASS:      cCustomer                                  *
*                                                          *
*   PURPOSE:    Overload "<<" operator.                    *
*                                                          *
*   RETURNS:    ostream&                                   *
*                                                          *
***********************************************************/
ostream& operator<< (ostream& output,
                     const cCustomer& cust_REF)
   {
   output << cust_REF._title     << ' '; 
   output << cust_REF._firstname << ' '; 
   output << cust_REF._lastname  << '\n'; 
   output << cust_REF._address1  << '\n';
   output << cust_REF._address2  << '\n';
   output << cust_REF._address3  << '\n';
   output << cust_REF._postcode;

   return output;
   } 


/***********************************************************
*                                                          *
*   A FUNCTION: Compare_Name()                             *
*                                                          *
*   CLASS:      cCustomer                                  *
*                                                          *
*   PURPOSE:    Compare given named and private data field *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    0 (int) if a match is found.               *
*                                                          *
***********************************************************/
int cCustomer::Compare_Name(char* name_PTR) const
   {
   return strcmp(name_PTR, _lastname);
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Compare_Area()                             *
*                                                          *
*   CLASS:      cCustomer                                  *
*                                                          *
*   PURPOSE:    Compare given area and private data field. *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    0 (int) if a match is found.               *
*                                                          *
***********************************************************/
int cCustomer::Compare_Area(char* Area_PTR) const
   {
   /* In the UK, a postcode consists of 2 groups of 3
    * letters and numbers: for example, "LE1 7RH".  Only the
    * first two letters of the first group are used in this
    * comparison.  These two letters (e.g. LE for Leicester,
    * DD for Dundee) give a sufficiently accurate area check
    * for our purposes.
    *
    * If you live outside the UK, apply an appropriate rule
    * for your own situation. */

   return strncmp(Area_PTR, _postcode, 2);
   }


/***********************************************************
************************************************************
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
************************************************************
***********************************************************/
class cCustomer_List  
   {
   public:
           cCustomer_List();  // Constructor
           ~cCustomer_List(); // Destructor  
           
      void Enter();
      void Name_Search() const;
      void Area_Search() const;
      void All_Search()  const;
      int  Menu() const;

   private:
      int        _max_list_size;
      cCustomer* _List_PTR;
      int        _list_size;
   };


/***********************************************************
*                                                          *
*   C FUNCTION: cCustomer_List()                           *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Load the customer file from disk into      *
*               a dynamically-created array.               *
*                                                          *
*   PRE:        None.                                      *
*   POST:       Detailed below.                            *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
cCustomer_List::cCustomer_List()
   {
   // No pre-conditions

   fstream file_ptr;
   _list_size = 0;
   _max_list_size = 0;
   
   // Try to open a data file on disk
   file_ptr.open("mail.dat",
                 ios::in | ios::nocreate | ios::binary);
   
   if (file_ptr.fail())
      {
      // Couldn't find an existing data file 
      cout << "WARNING: Can't find file 'mail.dat'\n";
      cout << "         A new file will be created.\n\n"; 
      }
   else
      {
      cout << "File 'mail.dat' opened successfully.\n";
         
      // Opened the existing data file successfully.
      // The first entry gives the number of records  
      file_ptr.read((char*) &_list_size, sizeof(_list_size));

      cout  << "The file presently contains " 
            << _list_size << " records.\n";
      }

   /* Memory for the database is to be allocated
    * dynamically.  Try to grab enough memory for the data
    * on disk (if any), plus 10 new records.   */
   _max_list_size = _list_size + 10;

   _List_PTR = new cCustomer[_max_list_size];

   if (_List_PTR == NULL)
      {
      cout << "Sufficient memory could not be allocated.";
      exit(1);
      }
   else
      {
      cout << "Memory allocated for "
           << _max_list_size << " records.\n\n";
      
      if (!file_ptr.fail())
         {
         // Read the records from disk (if poss)
         file_ptr.read((char*) _List_PTR, 
               _list_size*sizeof(cCustomer));
 
         if (file_ptr.fail())
            {
            cout << "A disk error was detected\n";
            exit(1);
            }
        
         file_ptr.close();
         }
      }

   // Post-conditions
   assert(_list_size >= 0);
   assert(_max_list_size >= 0);

   // end of function cCustomer_List()
   }


/***********************************************************
*                                                          *
*   D FUNCTION: ~cCustomer_List()                          *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Save the customer data on disk.            *
*               Return memory to free store.               *
*                                                          *
*   PRE:        Detailed below.                            *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
cCustomer_List::~cCustomer_List()
   {
   // Pre-conditions
   assert(_List_PTR != NULL);
   assert(_list_size >= 0);

   fstream file_ptr;
   
   file_ptr.open("mail.dat", ios::out | ios::binary);

   if (file_ptr.fail())
      {
      cout << "ERROR: Cannot create data file 'mail.dat'\n";
      }
   else
      {
      cout << "List size is " << _list_size << endl;
   
      // Write the file header first (see Load_Data())
      file_ptr.write((char*) &_list_size, sizeof(_list_size));
   
      //Now write the array of data
      file_ptr.write((char*) _List_PTR,
                     _list_size*sizeof(cCustomer));
   
      file_ptr.close();
   
      cout << "Data saved successfully.\n";
      }
      
   // No post-conditions
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Enter()                                    *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Enter records into database.               *
*                                                          *
*   PRE:        Detailed below.                            *
*   POST:       Detailed below.                            *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCustomer_List::Enter()
   {
   // Pre-conditions
   assert(_List_PTR != NULL);
   assert(_list_size >= 0);
   assert(_list_size <= _max_list_size);

   char answer[STRG_SZ];

   do
      {
      cin >> _List_PTR[_list_size];
              
      if (_List_PTR[_list_size].Compare_Name('\0'))
         {
         _list_size++;  // Only increment count if new data added
         }

      if (_list_size <= _max_list_size)
         {
         cout << "Do you want to enter further names? ";
         cin.getline(answer, STRG_SZ);
         }
      else
         {
         // List is full - set answer to 'n' to 
         // force termination of the loop. 
         cout << "Sorry - list is full.\n";
         answer[0] = 'n';
         }
      } while (answer[0] != 'n');   

   // Post-conditions
   assert(_list_size >= 0);
   assert(_list_size <= _max_list_size);
   }


/***********************************************************
*                                                          *
*   A FUNCTION: Name_Search()                              *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Search by customer name.                   *
*                                                          *
*   PRE:        Detailed below.                            *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCustomer_List::Name_Search() const
   {
   // Pre-conditions
   assert(_List_PTR != NULL);
   assert(_list_size >= 0);

   char name[STRG_SZ];
   int  i,  found;

   cout << "Last Name: ";
   cin.getline(name, STRG_SZ);

   found = 0;

   for(i=0; i<_list_size; i++)
      {
      if(!_List_PTR[i].Compare_Name(name))
         {
         cout << _List_PTR[i];
         found = 1;
         cout << "\n";
         }
      }

   if(!found)
      {
      cout << "No customers of that name found\n";
      }
   
   // No Post-condition checks
   }


/***********************************************************
*                                                          *
*   A FUNCTION: Area_Search()                              *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Search for customer by postcode.           *
*                                                          *
*   PRE:        Detailed below.                            *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCustomer_List::Area_Search() const
   {
   // Pre-conditions
   assert(_List_PTR != NULL);
   assert(_list_size >= 0);

   char area[STRG_SZ];
   int  i,  found;

   cout << "Postcode: ";
   cin.getline(area, STRG_SZ);

   found = 0;

   for(i=0; i<_list_size; i++)
      {
      if(!_List_PTR[i].Compare_Area(area))
         {
         cout << _List_PTR[i];
         found = 1;
         cout << "\n";
         }
      }

   if(!found)
      {
      cout << "No customers found in the area specified\n";
      }
   
   // No Post-condition checks
   }


/***********************************************************
*                                                          *
*   A FUNCTION: All_Search()                               *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Display all database records.              *
*                                                          *
*   PRE:        Detailed below.                            *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCustomer_List::All_Search() const
   {
   // Pre-conditions
   assert(_List_PTR != NULL);
   assert(_list_size >= 0);

   int  i,  found;

   found = 0;

   for(i=0; i<_list_size; i++)
      {
      cout << _List_PTR[i];
      found = 1;
      cout << "\n";
      }

   if(!found)
      {
      cout << "The list is empty.\n";
      }
   
   // No Post-condition checks
   }


/***********************************************************
*                                                          *
*   A FUNCTION: Menu()                                     *
*                                                          *
*   CLASS:      cCustomer_List                             *
*                                                          *
*   PURPOSE:    Tell user options and return valid choice. *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    User choice (int).                         *
*                                                          *
***********************************************************/
int cCustomer_List::Menu() const
   {
   int i;  
   char text_str[STRG_SZ];

   cout << " 0. Quit\n";
   cout << " 1. Enter New Data\n"; 
   
   if (_list_size)
      {
      cout << " 2. Search Database by Customer Name\n";
      cout << " 3. Search Database by Region\n";
      cout << " 4. Display all data\n"; 
      }
   
   int Valid_choice = 0;
   do 
      {
      cout << "\n? : ";
      cin.getline(text_str, STRG_SZ);
      i = text_str[0] - '0';
      
      if (_list_size)
         {
         Valid_choice = (i >= 0) && (i <= 4);
         }
      else  
         {
         Valid_choice = (i >= 0) && (i <= 1);
         } 
      } while(!Valid_choice);   
      
   return i;
   } 

int main() 
   { 
   cout << "PROTOTYPE MAILING LIST PROGRAM (O-O version)\n";
   cout << "--------------------------------------------\n\n";

   cCustomer_List The_List;
   int choice;

   do
      {
      choice = The_List.Menu();
          
      switch(choice) 
         {
         case 1: // Enter names  
            The_List.Enter();
            break;
         case 2: // Search by customer name  
            The_List.Name_Search(); 
            break;
         case 3: // Search by postcode  
            The_List.Area_Search(); 
            break;
         case 4: // "Search" for *all* records
            The_List.All_Search();   
         } 
      } while(choice);

   cout << "Finished.";

   return 0;   
   }

/***********************************************************
*                 *** END OF PROGRAM ***                   *
***********************************************************/