/***********************************************************
*                                                          *
*   A Bank Service Till (ATM) Framework (O-O version)      *
*                                                          *
*   Source is:     32_01.CPP                               *
*   Created:       07 December, 1994                       *
*   Last edited:   02 January, 1995                        *
*                                                          *
***********************************************************/

#include <iostream.h>
#include <stdlib.h>

// Maximum number of tries allowed to enter correct password
const int MAX_NUM_PASS_TRIES = 3;

// Various "flags"
enum eBool {FALSE, TRUE};  
enum eProcess {CHECK_CARD = 12, KEEP_CARD, EJECT_CARD};


/***********************************************************
************************************************************
*                                                          *
*   CLASS: cATM                                            *
*                                                          *
************************************************************
***********************************************************/
class cATM
   {
   public:
            cATM();
      eBool Hardware_OK();
      void  Check_supplies();
      void  Process_card(eProcess, int* = 0, int* = 0);
      void  Get_password(int&, int&, int);
      void  Deliver_services();
      void  Shutdown();
      int   Get_user_choice();

   private:
      void  _Order_cheque_book_from_bank();
      void  _Print_customer_balance();
      void  _Print_customer_mini_statement();
      void  _Dispense_cash_to_customer();

      eBool _cheque_OK;
      eBool _balance_OK;
      eBool _statement_OK;
      eBool _cash_OK;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cATM()                                     *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
***********************************************************/
cATM::cATM()
   {
   cout << "*** CONSTRUCTOR: cATM ***\n";
   }


/***********************************************************
*                                                          *
*   M FUNCTION: Hardware_OK()                              *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Polls hardware to determine system status. *
*                                                          *
***********************************************************/
eBool cATM::Hardware_OK()
   {
   cout << "*** cATM: WILL CHECK HARDWARE STATUS ***\n";

   // Say hardware is OK
   return TRUE;
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Check_supplies()                           *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Checks supplies of paper and cash.         *
*                                                          *
***********************************************************/
void cATM::Check_supplies()
   {
   cout << "*** cATM: WILL CHECK SUPPLIES ***\n";

   // Dummy values
   _cheque_OK = TRUE;
   _balance_OK = TRUE;
   _cash_OK = TRUE;
   _statement_OK = TRUE;
   } 


/***********************************************************
*                                                          *
*   M FUNCTION: Process_card()                             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Check, keep or eject cash card (simulated) *
*                                                          *
***********************************************************/
void cATM::Process_card(eProcess  required_op,
                        int*      acc_PTR,
                        int*      card_OK_PTR)
   { 
   /* NOTE: Process_card() is a function with a
      default argument.  Can call as:        
  
      Process_card(CHECK_CARD, card_OK);
      or
      Process_card(KEEP_CARD);
      Process_card(EJECT_CARD);  */
  
   switch (required_op)
      {
      case CHECK_CARD:
         // Using "extra" argument 
         cout << "Please insert your card.\n";
         cout << "*** SIMULATING INSERTION OF CORRECT CARD ***\n";
         // Card assumed OK
         cout << "Thank you.\n\n";
         *card_OK_PTR = TRUE;

         break;
                  
      case KEEP_CARD:
         cout << "Your card has been retained. ";
         cout << "Please contact your branch.\n\n";
         cout << "*** SIMULATING CARD RETAINED ***\n";
         break;
         
      case EJECT_CARD:
         cout << "Please take your card.\n\n";
         cout << "*** SIMULATING RETURN OF CARD ***\n";
         break;  
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Get_password()                             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Get password, and check it via BCL         *
*                                                          *
***********************************************************/
void cATM::Get_password(int& password_OK_REF,
                        int& tries_REF,
                        int account_no)
   {
   int password;
   
   tries_REF++;                                             
   cout << "Please enter your password : ";
   cin >> password;

   if (password == 1234)
      {
      cout << "*** PASSWORD IS CORRECT ***\n";
      password_OK_REF = TRUE;
      }
   else
      {
      password_OK_REF = FALSE;
      }
   
   if (!(password_OK_REF)) 
      {
      if (tries_REF < MAX_NUM_PASS_TRIES)
         {
         cerr << "Your password was incorrect.\n";
         // Recursive call to Get_password()          
         Get_password(password_OK_REF, tries_REF, account_no);
         }
      else
         {
         // Too many tries - could be a stolen card?  
         Process_card(KEEP_CARD);
         }
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Deliver_services()                         *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Deliver required customer services         *
*                                                          *
***********************************************************/
void cATM::Deliver_services()
   {
   int user_choice;
   
   do {
      user_choice = Get_user_choice();

      switch (user_choice)
         {  // If "Quit" then do nothing...
         case 1:
            _Order_cheque_book_from_bank();
         break;
      
         case 2:
            _Print_customer_balance();
         break;

         case 3:
            _Print_customer_mini_statement();
         break;

         case 4:
            _Dispense_cash_to_customer();
         break;
         }
      } while (user_choice != 0); // 0 is assumed quit
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Get_user_choice()                          *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Find out what user wants to do             *
*                                                          *
***********************************************************/
int cATM::Get_user_choice()
   {
   int service;

   cout << "\nAvailable options are: \n";
   cout << " 0. QUIT\n";
   if (_cheque_OK == TRUE)    cout << " 1. Cheque book\n";
   if (_balance_OK == TRUE)   cout << " 2. Balance\n";
   if (_statement_OK == TRUE) cout << " 3. Statement\n";
   if (_cash_OK == TRUE)      cout << " 4. Cash\n";
   cout << "\nPlease enter choice : ";

   // No checking yet...
   cin >> service;
 
   return service;
   } 

/***********************************************************
*                                                          *
*   M FUNCTION: _Order_cheque_book_from_bank()             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Order cheque book from bank.               *
*                                                          *
***********************************************************/
void cATM::_Order_cheque_book_from_bank()
   {
   cout << "\n*** SIMULATING ORDER FOR CHEQUE BOOK ***\n";
   cout << "Your cheque book will be posted to you.\n";

   _cheque_OK = FALSE;
   } 


/***********************************************************
*                                                          *
*   M FUNCTION: _Print_customer_balance()                  *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Print customer balance.                    *
*                                                          *
***********************************************************/
void cATM::_Print_customer_balance()
   {
   cout << "\n*** SIMULATED BALANCE  ***\n";
   cout << "Your balance is: 99.99\n";

   _balance_OK = FALSE;
   }
   

/***********************************************************
*                                                          *
*   M FUNCTION: _Print_customer_mini_statement()           *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Print customer mini statement              *
*                                                          *
***********************************************************/
void cATM::_Print_customer_mini_statement()
   {
   cout << "\n*** SIMULATED MINI STATEMENT ***\n";
   cout << "Please take your mini statement.\n";

   _statement_OK = FALSE;
   }


/***********************************************************
*                                                          *
*   M FUNCTION: _Dispense_cash_to_customer()               *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Dispense cash to customer                  *
*                                                          *
***********************************************************/
void cATM::_Dispense_cash_to_customer()
   {
   cout << "\n*** SIMULATED CASH DELIVERY ***\n";
   cout << "Please take your cash.\n";

   _cash_OK = FALSE;
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Shutdown()                                 *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Shut down cash machine.                    *
*                                                          *
***********************************************************/
void cATM::Shutdown()
   {
   cout << "\nPlease use another machine.\n";
   cout << "*** SIMULATING MACHINE SHUTDOWN ***\n";

   exit(1);
   }

/***********************************************************
***********************************************************/

int main()
   {
   cATM cash_machine;
   int acc;
   int valid_card_entered;  
   int tries, valid_password_entered;
   int user_wants_to_quit;

   // Simulate two customers in queue then manual shut down
   for (int customer = 1; customer <= 2; customer++)
      {
      cout << "\n----------------------------------\n";
      cout << "PROTOTYPE BANK AUTO TELLER MACHINE\n";
      cout << "----------------------------------\n\n";

      cash_machine.Check_supplies();

      cash_machine.Process_card(CHECK_CARD, &acc,
                                   &valid_card_entered);

      if (valid_card_entered)
         {
         tries = 0; 
         user_wants_to_quit = FALSE;

         cash_machine.Get_password(valid_password_entered,
                                      tries,acc);

         if (valid_password_entered)
            {
            // Keep delivering services till user
            // says quit (or hardware fault) 
            cash_machine.Deliver_services();
            cash_machine.Process_card(EJECT_CARD);
            }
         }
      } 

   // Simulate manual shutdown
   cash_machine.Shutdown();
             
   return 0;
   }

