/***********************************************************
*                                                          *
*   A Bank Service Till (ATM) Simulation (O-O version)     *
*                                                          *
*   Source is:     32_02.CPP                               *
*   Created:       20 Dec, 1994                            *
*   Last edited:   17 Mar, 1997                            *
*                                                          *
***********************************************************/

// Slightly tidier than version in Chapter 32
// MJP: 17 March, 1997

#include <iostream.h>

// extern "C" may be necessary on some compilers
// and won't do any harm on others
extern "C" 
{
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
}

// Width of monitor (in columns)
const int MON_SIZE = 30;

// Number of softkeys on monitor: 
const int NUM_SOFT_KEYS = 6;

// Maximum number of tries allowed to enter correct password
const int MAX_NUM_PASS_TRIES            = 3;

// Various "flags"
enum eBool {FALSE, TRUE};  
enum eProcess {CHECK_CARD = 12, KEEP_CARD, EJECT_CARD};


/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cBank_Computer_Link                            *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/

class cBank_Computer_Link
   {
   public:
            cBank_Computer_Link();

      eBool Password_OK(const int, const int) const;
      eBool Faulty() const;
      void  Alter_Balance(const int X) {_Balance += X;}
      int   Get_Balance() const {return _Balance;}  

   private:
      // Simulating only one single valid password
      // and account combination... 
      int   _Account;
      int   _Password;
      int   _Balance;
   };

/***********************************************************
*                                                          *
*   C FUNCTION:  cBank_Computer_Link()                     *
*                                                          *
*   CLASS:       cBank_Computer_Link                       *
*                                                          *
*   OVERVIEW:    Initiates connection with bank computer.  *
*                                                          *
*   PRE:         None.                                     *
*   POST:        None.                                     *
*                                                          *
***********************************************************/
cBank_Computer_Link::cBank_Computer_Link()
   {
   // In this prototype version, this constructor
   // simply provides basic information on one account.
   _Password = 1234;
   _Account  = 4321;
   _Balance  = 70;
   }


/***********************************************************
*                                                          *
*   A FUNCTION: Password_OK()                              *
*                                                          *
*   CLASS:      cBank_Computer_Link                        *
*                                                          *
*   OVERVIEW:   Passes password to bank computer.          *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       Got valid response.                        *
*                                                          *
*   RETURNS:    TRUE (eBool) if password correct,          *
*               FALSE otherwise.                           *
*                                                          *
***********************************************************/
eBool cBank_Computer_Link::Password_OK(const int PASS, 
                                       const int ACC) const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   // Simulate call to bank computer
   cerr << "\n*** SIMULATING CALL TO BANK COMPUTER ***\n";

   // Simulating one single valid password
   // and account combination... 
   if ((PASS == _Password) && (ACC == _Account))
      {
      cerr << "*** PASSWORD CORRECT ***\n";
      return TRUE;
      }
   else
      {
      cerr << "*** PASSWORD *NOT* CORRECT ***\n\n";
      return FALSE;
      }
   }


/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cBank_Computer_Link                        *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cBank_Computer_Link::Faulty() const
   {
   if ((rand() % 1000) == 0)
      {
      cerr << "*** SIMULATING BANK COMP LINK FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS: cCard_Reader                                    *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cCard_Reader
   {
   public:
            cCard_Reader(const eBool = FALSE);

      eBool Insert(int&);
      void  Eject();
      void  Keep();
      eBool Faulty() const;

   private:
      eBool _Card_present;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cCard_Reader()                             *
*                                                          *
*   CLASS:      cCard_Reader                               *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
***********************************************************/
cCard_Reader::cCard_Reader(const eBool CARD)
   {
   _Card_present = CARD;
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Insert()                                   *
*                                                          *
*   CLASS:      cCard_Reader                               *
*                                                          *
*   OVERVIEW:   Simulates insertion and reading of card.   *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if card readable, FALSE otherwise.    *
*                                                          *
***********************************************************/
eBool cCard_Reader::Insert(int& acc_REF)
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   if (_Card_present)
      {
      cerr << "*** THERE IS ALREADY A CARD IN THE MACHINE ***\n";
      }
   else
      {
      cerr << "*** SIMULATING INSERTION OF CARD ***\n";
      _Card_present = TRUE;

      cerr << "*** SIMULATING READING A/C NUMBER ***\n";
      acc_REF = 4321;
      }

   if ((rand() % 10) == 0)
      {
      cerr << "*** SIMULATING UNREADABLE CARD ***\n";
      return FALSE;
      }
   else
      {
      return TRUE;
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Eject()                                    *
*                                                          *
*   CLASS:      cCard_Reader                               *
*                                                          *
*   OVERVIEW:   Ejects card from machine (simulated)       *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCard_Reader::Eject()
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   if (!_Card_present)
      {
      cerr << "*** THERE IS NO CARD IN THE MACHINE ***\n";
      }
   else
      {
      cerr << "*** SIMULATING EJECTION OF CARD ***\n";
      _Card_present = FALSE;
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Keep()                                     *
*                                                          *
*   CLASS:      cCard_Reader                               *
*                                                          *
*   OVERVIEW:   Simulates retention of card by card reader *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cCard_Reader::Keep()
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   if (!_Card_present)
      {
      cerr << "*** THERE IS NO CARD IN THE MACHINE ***\n";
      }
   else
      {
      cerr << "*** SIMULATING RETENTION OF CARD ***\n";
      _Card_present = FALSE;
      }
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cCard_Reader                               *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cCard_Reader::Faulty() const
   {
   if ((rand() % 1000) == 0)
      {
      cerr << "*** SIMULATING CARD READER FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;   
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS: cKeypad                                         *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cKeypad
   {
   public:
            cKeypad(const int = 81); 
            
      int   Enter() const;
      eBool Faulty() const;

   private:
      int   _max_line_length;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cKeypad()                                  *
*                                                          *
*   CLASS:      cKeypad                                    *
*                                                          *
*   OVERVIEW:   Configures the keypad.                     *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
***********************************************************/
cKeypad::cKeypad(const int LENGTH)
   {
   _max_line_length = LENGTH; // Sets max input line length
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Enter()                                    *
*                                                          *
*   CLASS:      cKeypad                                    *
*                                                          *
*   OVERVIEW:   Simulates entry of data on keypad.         *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
int cKeypad::Enter() const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   char tmp[81];
   
   cin.getline(tmp, 81);
   
   return atoi(tmp);  
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cKeypad                                    *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cKeypad::Faulty() const
   {
   if ((rand() % 800) == 0)     
      {
      cerr << "*** SIMULATING KEYPAD FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS: cSoftKeys                                       *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cSoftKeys
   {
   public:
            cSoftKeys(const int = NUM_SOFT_KEYS);
      int   Key_Pressed() const;
      eBool Faulty() const;

   protected:
      int _Num_keys;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cSoftKeys                                  *
*                                                          *
*   CLASS:      cSoftKeys                                  *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
***********************************************************/
cSoftKeys::cSoftKeys(const int KEYS)
    {
    _Num_keys = KEYS;
    }


/***********************************************************
*                                                          *
*   A FUNCTION: Key_Pressed()                              *
*                                                          *
*   CLASS:      cSoftKey                                   *
*                                                          *
*   OVERVIEW:   Returns number of softkey pressed          *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    Key number.                                *
*                                                          *
***********************************************************/
int cSoftKeys::Key_Pressed() const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   char ip_STR[81]; // Input string buffer
   int  key_number;

   cin.getline(ip_STR,sizeof(ip_STR));
   key_number = atoi(ip_STR);

   return key_number;
   } 

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cSoftKeys                                  *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cSoftKeys::Faulty() const
   {
   if ((rand() % 1000) == 0)     
      {
      cerr << "*** SIMULATING SOFTKEY FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cCash_Dispenser                                *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cCash_Dispenser
   {
   public:
            cCash_Dispenser(const int=1000);
      int   Get_Cash_Level() const {return _Cash;}
      eBool Faulty() const;

   private:
      int   _Cash;  // Amount of cash available
   };

/***********************************************************
*                                                          *
*   C FUNCTION:  cCash_Dispenser()                         *
*                                                          *
*   CLASS:       cCash_Dispenser                           *
*                                                          *
*   PRE:         None.                                     *
*   POST:        None.                                     *
*                                                          *
***********************************************************/
cCash_Dispenser::cCash_Dispenser(const int INITIAL_CASH)
   {
   _Cash = INITIAL_CASH;
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cCash_Dispenser                            *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cCash_Dispenser::Faulty() const
   {
   if ((rand() % 100) == 0)
      {
      cerr << "*** SIMULATING CASH DISPENSER FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cOutput_Device                                 *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cOutput_Device
   {
   public:
              cOutput_Device(const int = 50, const int = 10);
      void    Display(char*) const;
      virtual eBool Faulty() const = 0; // Pure virtual

   private:
      int   _H_size;
      int   _V_size;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cOutput()                                  *
*                                                          *
*   CLASS:      cOutput_Device                             *
*                                                          *
*   OVERVIEW:   Configures output device.                  *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
***********************************************************/
cOutput_Device::cOutput_Device(const int HORIZ,
                               const int VERT)
   {
   // _H_size is horizontal size, in columns, of device
   _H_size = HORIZ;

   // _V_size is number of rows that can be displayed
   _V_size = VERT;
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Display()                                  *
*                                                          *
*   CLASS:      cOutput_Device                             *
*                                                          *
*   OVERVIEW:   Displays given message on output device.   *
*                                                          *
*   PRE:        string_PTR != 0                            *
*               Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cOutput_Device::Display(char* string_PTR) const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}
   // pre - string is not null
   assert(string_PTR != 0);

   // To change the display size requires only
   // local changes here.  
   char temp_str[80];
   strncpy(temp_str, string_PTR, _H_size);
   cout << temp_str;
   }


/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cVideo_Monitor                                 *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/

class cVideo_Monitor : virtual public cOutput_Device
   {
   public:
      void  Clear() const;
      eBool Faulty() const; 
      
   private:
   };


/***********************************************************
*                                                          *
*   A FUNCTION: Clear()                                    *
*                                                          *
*   CLASS:      cVideo_Monitor                             *
*                                                          *
*   OVERVIEW:   Clears video monitor display.              *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cVideo_Monitor::Clear() const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   // Print a number of blank lines to clear the
   // display: crude but (almost) universally effective

   // NOTE: disabled here to make book example clearer
   for (int line = 1; line <= 0; line++)
      {
      Display("\n");
      }
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cVideo_Monitor                             *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cVideo_Monitor::Faulty() const
   {
   if ((rand() % 10000) == 0)
      {
      cerr << "*** SIMULATING VIDEO MONITOR FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS: cPrinter                                        *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/

class cPrinter : virtual public cOutput_Device
   {
   public:
            cPrinter(const int = 10);

      void  Print_Mini();
      int   Get_Paper_Level() const {return _Paper_level;}
      eBool Faulty() const; 

   private:
      int   _Paper_level;
   };

/***********************************************************
*                                                          *
*   C FUNCTION: cPrinter()                                 *
*                                                          *
*   CLASS:      cPrinter                                   *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
***********************************************************/
cPrinter::cPrinter(const int P)
   {
   // Number of initial sheets of paper
   _Paper_level = P;
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Print_Mini()                               *
*                                                          *
*   CLASS:      cPrinter                                   *
*                                                          *
*   OVERVIEW:   Generates mini statement (simulated)       *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cPrinter::Print_Mini() 
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   cout << "*************************************\n";
   cout << "*     SIMULATED MINI STATEMENT      *\n";
   cout << "*************************************\n";
   cout << "\n  10\n  20\n  30\n  40\n";
   cout << "*************************************\n";

   _Paper_level--;
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cPrinter                                   *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cPrinter::Faulty() const
   {
   if ((rand() % 500) == 0)
      {
      cerr << "*** SIMULATING PRINTER FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cSoftKeyMonitor                                *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error.              *
*                                                          *
************************************************************
***********************************************************/
class cSoftKeyMonitor : public cSoftKeys,
                        public cVideo_Monitor
   {
   public:
      int   Find_Choice(char two_d[][MON_SIZE]) const; 
      eBool Faulty() const;
   };
                                     
/***********************************************************
*                                                          *
*   A FUNCTION: Find_Choice()                               *
*                                                          *
*   CLASS:      cSoftKeyMonitor                            *
*                                                          *
*   OVERVIEW:   Get user choice of service.                *
*                                                          *
*   PRE:        Hardware not faulty.                       *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
int cSoftKeyMonitor::Find_Choice(char two_d[][MON_SIZE]) const
   {
   // Pre - the hardware is OK 
   if (Faulty()) {throw 1;}

   int key;
   int key_OK;
   int valid_entry[NUM_SOFT_KEYS];  

   Display("\nAvailable options: \n");

   for (int x = 0; x < _Num_keys; x++)
       {
       valid_entry[x] = 0; // Default
          
       if (two_d[x][0] != 'X')
          {
          Display(&two_d[x][0]);
          Display("\n");
          valid_entry[x] = 1;
          }
       }     

     do
      {
      cerr << "Please enter the appropriate number : ";
      key = Key_Pressed();
      
      // Determine if key choice is valid 
      if ((key >= 0) && (key < _Num_keys))
         {
         key_OK = valid_entry[key];
         }
      else
         {
         key_OK = 0; 
         }
       } while (!key_OK);
 
   Display("Thank you.\n\n");

   return key;
   }

/***********************************************************
*                                                          *
*   A FUNCTION: Faulty()                                   *
*                                                          *
*   CLASS:      cSoftKeyMonitor                            *
*                                                          *
*   OVERVIEW:   Detects hardware fault.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    TRUE if fault detected, FALSE otherwise.   *
*                                                          *
***********************************************************/
eBool cSoftKeyMonitor::Faulty() const
   { 
   if (cSoftKeys::Faulty() || cVideo_Monitor::Faulty())
      {
      cerr << "*** SIMULATING SOFTKEY MONITOR FAULT ***\n";
      return TRUE;
      }
   else
      {
      return FALSE;
      }
   }

/***********************************************************
************************************************************
*                                                          *
*   CLASS:  cATM                                           *
*                                                          *
*   EXCEPT: Throws 1 (int) on hardware error               *
*                                                          *
************************************************************
***********************************************************/
class cATM
   {
   public:
      eBool Hardware_OK();
      void  Check_supplies();
      void  Process_card(eProcess, int* = 0, int* = 0);
      void  Get_Password(int&, int&, int);
      void  Deliver_services();
      void  Shutdown(int);
      int   Get_user_choice();

   private:
      void  _Order_cheque_book_from_bank();
      void  _Print_customer_Balance();
      void  _Print_customer_mini_statement();
      void  _Dispense_Cash_to_customer();

      eBool _cheque_OK;
      eBool _Balance_OK;
      eBool _statement_OK;
      eBool _Cash_OK;

      cKeypad             _Kp;
      cSoftKeyMonitor     _SKM;
      cCard_Reader        _CR;
      cCash_Dispenser     _CD;
      cBank_Computer_Link _BCL;
      cPrinter            _Pr;
   };

/***********************************************************
*                                                          *
*   M FUNCTION: Hardware_OK()                              *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Polls hardware to determine system status. *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
eBool cATM::Hardware_OK()
   {
   // Check all hardware    
   if (_Kp.Faulty() || _SKM.Faulty() || _CR.Faulty() ||
       _CD.Faulty() || _BCL.Faulty() || _Pr.Faulty())
      {
      return FALSE;  
      }
   else
      {
      return TRUE;
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Check_supplies()                           *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Checks supplies of paper and cash.         *
*                                                          *
*   PRE:        All hardware working normally              *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::Check_supplies()
   {
   // Do full check of all hardware before new customer
   // uses the machine
   if (!Hardware_OK()) {throw 1;} 

   _SKM.Clear();
   _SKM.Display("Next customer please wait...\n\n");


   // Reset for new customer
   _cheque_OK = TRUE;

   if (_CD.Get_Cash_Level() > 0)  
      {
      // Reset these for new customer
      _Balance_OK = TRUE;
      _Cash_OK = TRUE;
      }
   else   
      {
      _Balance_OK = FALSE;
      _Cash_OK = FALSE; 
      }

   if (_Pr.Get_Paper_Level() > 0)    
      {
      // Reset these for new customer
      _Balance_OK = TRUE;
      _statement_OK = TRUE;
      }
   else   
      {
      _Balance_OK = FALSE;
      _statement_OK = FALSE;
      } 
   } 


/***********************************************************
*                                                          *
*   M FUNCTION: Process_card()                             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Check, keep or eject cash card (simulated) *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::Process_card(eProcess  required_op,
                        int*      acc_PTR,
                        int*      card_OK_PTR)
   { 
   /* NOTE: Process_card() is a function with a
      default argument.  Can call as:        
  
      Process_card(CHECK_CARD, card_OK);
      or
      Process_card(KEEP_CARD);
      Process_card(EJECT_CARD);  */
  
   switch (required_op)
      {
      case CHECK_CARD:
         // Using "extra" argument 
         _SKM.Clear();
         _SKM.Display("Please insert your card.\n\n");

         if (_CR.Insert(*acc_PTR))
            { // Valid card inserted
            _SKM.Display("\nThank you.\n\n");
            *card_OK_PTR = TRUE;
            }
         else
            {
            _SKM.Display("Sorry - your card is unreadable.\n\n");
            _CR.Eject();
            *card_OK_PTR = FALSE;
            }
         break;
                  
      case KEEP_CARD:
         _SKM.Display("Your card has been retained. ");
         _SKM.Display("Please contact your branch.\n\n");
         _CR.Keep();
         break;
         
      case EJECT_CARD:
         _SKM.Display("Please take your card.\n\n");
         _CR.Eject();
         break;  
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Get_Password()                             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Get password, and check it via BCL         *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::Get_Password(int& password_OK_REF,
                        int& tries_REF,
                        int account_no)
   {
   int password;
   
   tries_REF++;                                             
   _SKM.Display("Please enter your password on the keypad : ");
   password = _Kp.Enter();

   if (_BCL.Password_OK(password, account_no))
      {
      password_OK_REF = TRUE;
      }
   else
      {
      password_OK_REF = FALSE;
      }
   
   if (!(password_OK_REF)) 
      {
      if (tries_REF < MAX_NUM_PASS_TRIES)
         {
         cerr << "Your password was incorrect.\n";
         // Recursive call to Get_Password()          
         Get_Password(password_OK_REF, tries_REF, account_no);
         }
      else
         {
         // Too many tries - could be a stolen card?  
         Process_card(KEEP_CARD);
         }
      }
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Deliver_services()                         *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Deliver required customer services         *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::Deliver_services()
   {
   int user_choice;
   
   do {
      user_choice = Get_user_choice();

      switch (user_choice)
         {  // If "Quit" then do nothing...
         case 1:
            _Order_cheque_book_from_bank();
         break;
      
         case 2:
            _Print_customer_Balance();
         break;

         case 3:
            _Print_customer_mini_statement();
         break;

         case 4:
            _Dispense_Cash_to_customer();
         break;
         }
      } while (user_choice != 0); // 0 is assumed quit
   }

/***********************************************************
*                                                          *
*   M FUNCTION: Get_user_choice()                          *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Find out what user wants to do             *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
int cATM::Get_user_choice()
   {
   int service;

   char options[NUM_SOFT_KEYS][MON_SIZE] = {{" 0. Quit"},     
                                                   {"X"},
                                                   {"X"},
                                                   {"X"},
                                                   {"X"},
                                                   {"X"}};

   if (_cheque_OK == TRUE)    strcpy(options[1], " 1. Cheque book");
   if (_Balance_OK == TRUE)   strcpy(options[2], " 2. Balance");
   if (_statement_OK == TRUE) strcpy(options[3], " 3. Statement");
   if (_Cash_OK == TRUE)      strcpy(options[4], " 4. Cash");  

   service = _SKM.Find_Choice(options);
 
   return service;
   } 

/***********************************************************
*                                                          *
*   M FUNCTION: _Order_cheque_book_from_bank()             *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Order cheque book from bank.               *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::_Order_cheque_book_from_bank()
   {
   cerr << "*** SIMULATING ORDER FOR CHEQUE BOOK ***\n\n";
   _SKM.Display("Your cheque book will be posted to you.\n");

   _cheque_OK = FALSE;
   } 


/***********************************************************
*                                                          *
*   M FUNCTION: _Print_customer_Balance()                  *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Print customer balance.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::_Print_customer_Balance()
   {
   char tmp[12];

   // String conversion 
   sprintf(tmp, "%d", _BCL.Get_Balance());

   _Pr.Display("Your balance is: "); 
   _Pr.Display(tmp);
   _Pr.Display("\n");

   _Balance_OK = FALSE;
   }
   

/***********************************************************
*                                                          *
*   M FUNCTION: _Print_customer_mini_statement()           *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Print customer mini statement              *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::_Print_customer_mini_statement()
   {
   _Pr.Print_Mini();
   _SKM.Display("\nPlease take your mini statement.\n");

   _statement_OK = FALSE;
   }
   

/***********************************************************
*                                                          *
*   M FUNCTION: _Dispense_Cash_to_customer()               *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Dispense cash to customer                  *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::_Dispense_Cash_to_customer()
   {
   int service, request;

   char options[NUM_SOFT_KEYS][MON_SIZE] = {{" 0. QUIT"},     
                                             {" 1. 10"},
                                             {" 2. 20"},
                                             {" 3. 50"},
                                            {" 4. 100"},
                               {" 5. Some other amount"}};

   service = _SKM.Find_Choice(options);

   switch (service)
      {
      case 0:
         request = 0;
      break;
   
      case 1:
         request = 10;
      break;
      
      case 2:
         request = 20;
      break;

      case 3:
         request = 50;
      break;

      case 4:
         request = 100;
      break;

      case 5:
         request = -1;
      break;
      }

   if (request > 0)
      {
      if (request <= _BCL.Get_Balance())
         {
         _BCL.Alter_Balance(-request);

         cerr << "*** SIMULATING CASH DELIVERY ***\n\n";
         _SKM.Display("Please take your cash.\n");
         _Cash_OK = FALSE;
         }
      else
         { 
         _SKM.Display("Sorry - the requested amount ");
         _SKM.Display("exceeds your credit limit.\n");
         }
      }
   else   
      {
      if (request < 0)
         {
         _SKM.Display("How much cash do you require? ");
         request = _Kp.Enter();  

         if (request % 10 != 0)  // Only multiples of 10...
            {
            _SKM.Display("Sorry - must be multiple of 10\n");
            }
         else
            { 
            if (request <= _BCL.Get_Balance())
               {
               _BCL.Alter_Balance(-request);

               cerr << "*** SIMULATING CASH DELIVERY ***\n\n";
               _SKM.Display("Please take your cash.\n");
               _Cash_OK = FALSE;
               }
            else
               { 
               _SKM.Display("Sorry - the requested amount ");
               _SKM.Display("exceeds your credit limit. ");
               }
            }
          }    
       } 
    }

/***********************************************************
*                                                          *
*   M FUNCTION: Shutdown()                                 *
*                                                          *
*   CLASS:      cATM                                       *
*                                                          *
*   OVERVIEW:   Shut down cash machine.                    *
*                                                          *
*   PRE:        None.                                      *
*   POST:       None.                                      *
*                                                          *
*   RETURNS:    void.                                      *
*                                                          *
***********************************************************/
void cATM::Shutdown(int level)
   {
   if (level)
      {
      _SKM.Display("A hardware error has been detected.\n\n");
      cerr << "*** SIMULATING EMERGENCY SHUTDOWN ***\n";
      }
   else
      {   
      _SKM.Display("Please use another machine.\n\n");
      cerr << "*** SIMULATING ORDINARY MACHINE SHUTDOWN ***\n";
      } 
   
   exit(1);
   }

/***********************************************************
***********************************************************/

int main()
   {
   cATM Cash_machine;
   int acc;
   int valid_card_entered;  
   int tries, valid_Password_entered;
   int user_wants_to_quit;

   // Set up non-repeatable random numbers for simulation
   srand((unsigned)time(NULL));

   try
      {
      // Simulate (up to) five customers in queue
      // then manual shut down 
      for (int customer = 1; customer <= 5; customer++)
         {
         cout << "\n----------------------------------\n";
         cout << "PROTOTYPE BANK AUTO TELLER MACHINE\n";
         cout << "----------------------------------\n\n";

         Cash_machine.Check_supplies();

         Cash_machine.Process_card(CHECK_CARD, &acc,
                                   &valid_card_entered);

         if (valid_card_entered)
            {
            tries = 0; 
            user_wants_to_quit = FALSE;

            Cash_machine.Get_Password(valid_Password_entered,
                                      tries,acc);

            if (valid_Password_entered)
               {
               // Keep delivering services till user
               // says quit (or hardware fault) 
               Cash_machine.Deliver_services();
               Cash_machine.Process_card(EJECT_CARD);
               }
            }
         } 

         // Simulate manual shutdown
         Cash_machine.Shutdown(0);
         } // End of try block

   // Deal with hardware or other errors
   catch (...)
      {
      Cash_machine.Shutdown(1);
      try
         {
         // Keep card: customer may have caused fault...
         Cash_machine.Process_card(KEEP_CARD);
         exit(1);
         }
      catch(...)
         {
         exit(1);
         }       
      }
          
   return 0;
   }
